import typing
from typing_extensions import TypedDict, Literal
from datetime import datetime

from opentrons_shared_data.pipette.dev_types import LabwareUri

from .types import AttitudeMatrix, PipetteOffset, SourceType


class CalibrationStatusDict(TypedDict):
    markedBad: bool
    source: typing.Optional[str]
    markedAt: typing.Optional[datetime]


class TipLengthCalibration(TypedDict):
    tipLength: float
    lastModified: datetime
    source: SourceType
    status: CalibrationStatusDict
    uri: typing.Union[LabwareUri, Literal[""]]


class ModuleDict(TypedDict):
    parent: str
    fullParent: str


class CalibrationIndexDict(TypedDict):
    """
    The dict that is returned from
    the index.json file.
    """

    uri: str
    slot: str
    module: ModuleDict


class OffsetDict(TypedDict):
    offset: typing.List[float]
    lastModified: datetime


class TipLengthDict(TypedDict):
    length: float
    lastModified: datetime


class CalibrationDict(TypedDict):
    """
    The dict that is returned from a labware
    offset file.
    """

    default: OffsetDict
    tipLength: TipLengthDict


class PipetteCalibrationData(TypedDict):
    offset: PipetteOffset
    tiprack: str
    uri: str
    last_modified: datetime
    source: SourceType
    status: CalibrationStatusDict


class DeckCalibrationData(TypedDict):
    attitude: AttitudeMatrix
    last_modified: datetime
    source: SourceType
    pipette_calibrated_with: typing.Optional[str]
    tiprack: typing.Optional[str]
    status: CalibrationStatusDict


PipTipLengthCalibration = typing.Dict[str, TipLengthCalibration]



"""Geometry state getters."""
from dataclasses import dataclass
from typing import Optional


from opentrons.types import Point
from opentrons.hardware_control.dev_types import PipetteDict

from .. import errors
from ..types import LoadedLabware, WellLocation, WellOrigin
from .labware import LabwareView


DEFAULT_TIP_DROP_HEIGHT_FACTOR = 0.5

# TODO(mc, 2020-11-12): reconcile this data structure with WellGeometry


@dataclass(frozen=True)
class TipGeometry:
    """Tip geometry data."""

    effective_length: float
    diameter: float
    volume: int


# TODO(mc, 2021-06-03): continue evaluation of which selectors should go here
# vs which selectors should be in LabwareView
class GeometryView:
    """Geometry computed state getters."""

    _labware: LabwareView

    def __init__(self, labware_view: LabwareView) -> None:
        """Initialize a GeometryView instance."""
        self._labware = labware_view

    def get_labware_highest_z(self, labware_id: str) -> float:
        """Get the highest Z-point of a labware."""
        labware_data = self._labware.get(labware_id)

        return self._get_highest_z_from_labware_data(labware_data)

    def get_all_labware_highest_z(self) -> float:
        """Get the highest Z-point across all labware."""
        return max(
            [
                self._get_highest_z_from_labware_data(lw_data)
                for lw_data in self._labware.get_all()
            ]
        )

    def get_labware_parent_position(self, labware_id: str) -> Point:
        """Get the position of the labware's parent slot (deck or module)."""
        labware_data = self._labware.get(labware_id)
        slot_pos = self._labware.get_slot_position(labware_data.location.slot)

        return slot_pos

    def get_labware_origin_position(self, labware_id: str) -> Point:
        """Get the position of the labware's origin, without calibration."""
        labware_data = self._labware.get(labware_id)
        slot_pos = self._labware.get_slot_position(labware_data.location.slot)
        origin_offset = self._labware.get_definition(labware_id).cornerOffsetFromSlot

        return Point(
            x=slot_pos.x + origin_offset.x,
            y=slot_pos.y + origin_offset.y,
            z=slot_pos.z + origin_offset.z,
        )

    def get_labware_position(self, labware_id: str) -> Point:
        """Get the calibrated origin of the labware."""
        origin_pos = self.get_labware_origin_position(labware_id)
        cal_offset = self._labware.get_calibration_offset(labware_id)

        return Point(
            x=origin_pos.x + cal_offset.x,
            y=origin_pos.y + cal_offset.y,
            z=origin_pos.z + cal_offset.z,
        )

    def get_well_position(
        self,
        labware_id: str,
        well_name: str,
        well_location: Optional[WellLocation] = None,
    ) -> Point:
        """Get the absolute position of a well in a labware."""
        labware_pos = self.get_labware_position(labware_id)
        well_def = self._labware.get_well_definition(labware_id, well_name)
        well_depth = well_def.depth

        if well_location is not None:
            offset = well_location.offset

            if well_location.origin == WellOrigin.TOP:
                offset = (offset[0], offset[1], offset[2] + well_depth)

        else:
            offset = (0, 0, well_depth)

        return Point(
            x=labware_pos.x + offset[0] + well_def.x,
            y=labware_pos.y + offset[1] + well_def.y,
            z=labware_pos.z + offset[2] + well_def.z,
        )

    def _get_highest_z_from_labware_data(self, lw_data: LoadedLabware) -> float:
        labware_pos = self.get_labware_position(lw_data.id)
        definition = self._labware.get_definition(lw_data.id)
        z_dim = definition.dimensions.zDimension

        return labware_pos.z + z_dim

    # TODO(mc, 2020-11-12): reconcile with existing protocol logic and include
    # data from tip-length calibration once v4.0.0 is in `edge`
    def get_effective_tip_length(
        self,
        labware_id: str,
        pipette_config: PipetteDict,
    ) -> float:
        """Given a labware and a pipette's config, get the effective tip length.

        Effective tip length is the nominal tip length less the distance the
        tip overlaps with the pipette nozzle.
        """
        labware_uri = self._labware.get_definition_uri(labware_id)
        nominal_length = self._labware.get_tip_length(labware_id)
        overlap_config = pipette_config["tip_overlap"]
        default_overlap = overlap_config.get("default", 0)
        overlap = overlap_config.get(labware_uri, default_overlap)

        return nominal_length - overlap

    # TODO(mc, 2020-11-12): reconcile with existing geometry logic
    def get_tip_geometry(
        self,
        labware_id: str,
        well_name: str,
        pipette_config: PipetteDict,
    ) -> TipGeometry:
        """Given a labware, well, and hardware pipette config, get the tip geometry.

        Tip geometry includes effective tip length, tip diameter, and tip volume,
        which is all data required by the hardware controller for proper tip handling.
        """
        effective_length = self.get_effective_tip_length(labware_id, pipette_config)
        well_def = self._labware.get_well_definition(labware_id, well_name)

        if well_def.shape != "circular":
            raise errors.LabwareIsNotTipRackError(
                f"Well {well_name} in labware {labware_id} is not circular."
            )

        return TipGeometry(
            effective_length=effective_length,
            diameter=well_def.diameter,  # type: ignore[arg-type]
            # TODO(mc, 2020-11-12): WellDefinition type says totalLiquidVolume
            #  is a float, but hardware controller expects an int
            volume=int(well_def.totalLiquidVolume),
        )

    # TODO(mc, 2020-11-12): support pre-PAPIv2.2/2.3 behavior of dropping the tip
    # 10mm above well bottom
    def get_tip_drop_location(
        self,
        labware_id: str,
        pipette_config: PipetteDict,
    ) -> WellLocation:
        """Get tip drop location given labware and hardware pipette."""
        # return to top if labware is fixed trash
        is_fixed_trash = self._labware.get_has_quirk(
            labware_id=labware_id,
            quirk="fixedTrash",
        )

        if is_fixed_trash:
            return WellLocation()

        nominal_length = self._labware.get_tip_length(labware_id)
        offset_factor = pipette_config["return_tip_height"]

        return WellLocation(offset=(0, 0, -nominal_length * offset_factor))



"""Protocol engine module.

The protocol_engine module contains the logic necessary to take a stream of
protocol commands, issued by some arbitrary protocol runner, and turn it into
protocol state and side-effects like robot movements.
"""

from .create_protocol_engine import create_protocol_engine
from .protocol_engine import ProtocolEngine
from .errors import ProtocolEngineError
from .commands import Command, CommandRequest, CommandStatus, CommandType
from .state import State, StateView
from .types import (
    CalibrationOffset,
    DeckSlotLocation,
    Dimensions,
    EngineStatus,
    LabwareLocation,
    LoadedLabware,
    LoadedPipette,
    PipetteName,
    WellLocation,
    WellOrigin,
)

__all__ = [
    # main factory and interface exports
    "create_protocol_engine",
    "ProtocolEngine",
    # error types
    "ProtocolEngineError",
    # top level command unions and values
    "Command",
    "CommandRequest",
    "CommandStatus",
    "CommandType",
    # state interfaces and models
    "State",
    "StateView",
    # public value interfaces and models
    "CalibrationOffset",
    "DeckSlotLocation",
    "Dimensions",
    "EngineStatus",
    "LabwareLocation",
    "LoadedLabware",
    "LoadedPipette",
    "PipetteName",
    "WellLocation",
    "WellOrigin",
]



from typing import Union, Sequence, List, Any

from opentrons.protocol_api.labware import Well
from opentrons.types import Location


CommandLocation = Union[Location, None, Sequence, Well]


def listify(location: Any) -> List:
    if isinstance(location, list):
        try:
            return listify(location[0])
        except IndexError:
            return [location]
    else:
        return [location]


def _stringify_new_loc(loc: Union[Location, Well]) -> str:
    if isinstance(loc, Location):
        if loc.labware.is_empty:
            return str(loc.point)
        else:
            return repr(loc.labware)
    elif isinstance(loc, Well):
        return str(loc)
    else:
        raise TypeError(loc)


def stringify_location(location: CommandLocation) -> str:
    loc_str_list = [_stringify_new_loc(loc) for loc in listify(location)]
    return ", ".join(loc_str_list)



from __future__ import annotations
from dataclasses import dataclass
import logging
import json
import numbers
from typing import Any, Dict, List, Mapping, Union, Tuple, Sequence, TYPE_CHECKING

from opentrons import config
from opentrons.config import feature_flags as ff
from opentrons_shared_data.pipette import model_config, name_config, fuse_specs

if TYPE_CHECKING:
    from opentrons_shared_data.pipette.dev_types import (
        PipetteName,
        PipetteModel,
        UlPerMm,
        Quirk,
        PipetteFusedSpec,
        LabwareUri,
    )


log = logging.getLogger(__name__)


@dataclass(frozen=True)
class PipetteConfig:
    top: float
    bottom: float
    blow_out: float
    drop_tip: float
    pick_up_current: float
    pick_up_distance: float
    pick_up_increment: float
    pick_up_presses: int
    pick_up_speed: float
    aspirate_flow_rate: float
    dispense_flow_rate: float
    channels: float
    nozzle_offset: Tuple[float, float, float]
    plunger_current: float
    drop_tip_current: float
    drop_tip_speed: float
    min_volume: float
    max_volume: float
    ul_per_mm: UlPerMm
    quirks: List[Quirk]
    tip_length: float  # TODO(seth): remove
    # TODO: Replace entirely with tip length calibration
    tip_overlap: Dict[str, float]
    display_name: str
    name: PipetteName
    back_compat_names: List[PipetteName]
    return_tip_height: float
    blow_out_flow_rate: float
    max_travel: float
    home_position: float
    steps_per_mm: float
    idle_current: float
    default_blow_out_flow_rates: Dict[str, float]
    default_aspirate_flow_rates: Dict[str, float]
    default_dispense_flow_rates: Dict[str, float]
    model: PipetteModel
    default_tipracks: List[LabwareUri]


# Notes:
# - multi-channel pipettes share the same dimensional offsets
# - single-channel pipettes have different lengths
# - Default number of seconds to aspirate/dispense a pipette's full volume,
#     and these times were chosen to mimic normal human-pipetting motions.
#     However, accurate speeds are dependent on environment (ex: liquid
#     viscosity), therefore a pipette's flow-rates (ul/sec) should be set by
#     protocol writer

# Multi-channel y offset caluclations:
DISTANCE_BETWEEN_NOZZLES = 9
NUM_MULTI_CHANNEL_NOZZLES = 8
MULTI_LENGTH = (NUM_MULTI_CHANNEL_NOZZLES - 1) * DISTANCE_BETWEEN_NOZZLES
Y_OFFSET_MULTI = MULTI_LENGTH / 2
Z_OFFSET_MULTI = -25.8

Z_OFFSET_P10 = -13  # longest single-channel pipette
Z_OFFSET_P50 = 0
Z_OFFSET_P300 = 0
Z_OFFSET_P1000 = 20  # shortest single-channel pipette

NOZZLE_OFFSET_DEFAULT = [0.0, 0.0, 0.0]

LOW_CURRENT_DEFAULT = 0.05

config_models = list(model_config()["config"].keys())
config_names = list(name_config().keys())
configs = model_config()["config"]
#: A list of pipette model names for which we have config entries
MUTABLE_CONFIGS = model_config()["mutableConfigs"]
#: A list of mutable configs for pipettes
VALID_QUIRKS = model_config()["validQuirks"]
#: A list of valid quirks for pipettes


def load(pipette_model: PipetteModel, pipette_id: str = None) -> PipetteConfig:
    """
    Load pipette config data


    This function loads from a combination of

    - the pipetteModelSpecs.json file in the wheel (should never be edited)
    - the pipetteNameSpecs.json file in the wheel (should never be edited)
    - any config overrides found in
      ``opentrons.config.CONFIG['pipette_config_overrides_dir']``

    This function reads from disk each time, so changes to the overrides
    will be picked up in subsequent calls.

    :param str pipette_model: The pipette model name (i.e. "p10_single_v1.3")
                              for which to load configuration
    :param pipette_id: An (optional) unique ID for the pipette to locate
                       config overrides. If the ID is not specified, the system
                       assumes this is a simulated pipette and does not
                       save settings. If the ID is specified but no overrides
                       corresponding to the ID are found, the system creates a
                       new overrides file for it.
    :type pipette_id: str or None
    :raises KeyError: if ``pipette_model`` is not in the top-level keys of
                      the pipetteModelSpecs.json file (and therefore not in
                      :py:attr:`configs`)

    :returns PipetteConfig: The configuration, loaded and checked
    """

    # Load the model config and update with the name config
    cfg = fuse_specs(pipette_model)

    # Load overrides if we have a pipette id
    if pipette_id:
        try:
            override = load_overrides(pipette_id)
            if "quirks" in override.keys():
                override["quirks"] = [
                    qname for qname, qval in override["quirks"].items() if qval
                ]
            for legacy_key in (
                "defaultAspirateFlowRate",
                "defaultDispenseFlowRate",
                "defaultBlowOutFlowRate",
            ):
                override.pop(legacy_key, None)

        except FileNotFoundError:
            save_overrides(pipette_id, {}, pipette_model)
            log.info(
                "Save defaults for pipette model {} and id {}".format(
                    pipette_model, pipette_id
                )
            )
        else:
            cfg.update(override)  # type: ignore

    # the ulPerMm functions are structured in pipetteModelSpecs.json as
    # a list sorted from oldest to newest. That means the latest functions
    # are always the last element and, as of right now, the older ones are
    # the first element (for models that only have one function, the first
    # and last elements are the same, which is fine). If we add more in the
    # future, we’ll have to change this code to select items more
    # intelligently
    if ff.use_old_aspiration_functions():
        log.debug("Using old aspiration functions")
        ul_per_mm = cfg["ulPerMm"][0]
    else:
        ul_per_mm = cfg["ulPerMm"][-1]

    smoothie_configs = cfg["smoothieConfigs"]
    res = PipetteConfig(
        top=ensure_value(cfg, "top", MUTABLE_CONFIGS),
        bottom=ensure_value(cfg, "bottom", MUTABLE_CONFIGS),
        blow_out=ensure_value(cfg, "blowout", MUTABLE_CONFIGS),
        drop_tip=ensure_value(cfg, "dropTip", MUTABLE_CONFIGS),
        pick_up_current=ensure_value(cfg, "pickUpCurrent", MUTABLE_CONFIGS),
        pick_up_distance=ensure_value(cfg, "pickUpDistance", MUTABLE_CONFIGS),
        pick_up_increment=ensure_value(cfg, "pickUpIncrement", MUTABLE_CONFIGS),
        pick_up_presses=ensure_value(cfg, "pickUpPresses", MUTABLE_CONFIGS),
        pick_up_speed=ensure_value(cfg, "pickUpSpeed", MUTABLE_CONFIGS),
        aspirate_flow_rate=cfg["defaultAspirateFlowRate"]["value"],
        dispense_flow_rate=cfg["defaultDispenseFlowRate"]["value"],
        channels=ensure_value(cfg, "channels", MUTABLE_CONFIGS),
        nozzle_offset=cfg.get("nozzleOffset", NOZZLE_OFFSET_DEFAULT),  # type: ignore
        plunger_current=ensure_value(cfg, "plungerCurrent", MUTABLE_CONFIGS),
        drop_tip_current=ensure_value(cfg, "dropTipCurrent", MUTABLE_CONFIGS),
        drop_tip_speed=ensure_value(cfg, "dropTipSpeed", MUTABLE_CONFIGS),
        min_volume=ensure_value(cfg, "minVolume", MUTABLE_CONFIGS),
        max_volume=ensure_value(cfg, "maxVolume", MUTABLE_CONFIGS),
        ul_per_mm=ul_per_mm,
        quirks=validate_quirks(ensure_value(cfg, "quirks", MUTABLE_CONFIGS)),
        tip_overlap=cfg["tipOverlap"],
        tip_length=ensure_value(cfg, "tipLength", MUTABLE_CONFIGS),
        display_name=ensure_value(cfg, "displayName", MUTABLE_CONFIGS),
        name=cfg["name"],
        back_compat_names=cfg.get("backCompatNames", []),
        return_tip_height=cfg.get("returnTipHeight", 0.5),
        blow_out_flow_rate=cfg["defaultBlowOutFlowRate"]["value"],
        max_travel=smoothie_configs["travelDistance"],
        home_position=smoothie_configs["homePosition"],
        steps_per_mm=smoothie_configs["stepsPerMM"],
        idle_current=cfg.get("idleCurrent", LOW_CURRENT_DEFAULT),
        default_blow_out_flow_rates=cfg["defaultBlowOutFlowRate"].get(
            "valuesByApiLevel", {"2.0": cfg["defaultBlowOutFlowRate"]["value"]}
        ),
        default_dispense_flow_rates=cfg["defaultDispenseFlowRate"].get(
            "valuesByApiLevel", {"2.0": cfg["defaultDispenseFlowRate"]["value"]}
        ),
        default_aspirate_flow_rates=cfg["defaultAspirateFlowRate"].get(
            "valuesByApiLevel", {"2.0": cfg["defaultAspirateFlowRate"]["value"]}
        ),
        model=pipette_model,
        default_tipracks=cfg["defaultTipracks"],
    )

    return res


def piecewise_volume_conversion(ul: float, sequence: List[List[float]]) -> float:
    """
    Takes a volume in microliters and a sequence representing a piecewise
    function for the slope and y-intercept of a ul/mm function, where each
    sub-list in the sequence contains:

      - the max volume for the piece of the function (minimum implied from the
        max of the previous item or 0
      - the slope of the segment
      - the y-intercept of the segment

    :return: the ul/mm value for the specified volume
    """
    # pick the first item from the seq for which the target is less than
    # the bracketing element
    for x in sequence:
        if ul <= x[0]:
            # use that element to calculate the movement distance in mm
            return x[1] * ul + x[2]

    # Compatibility with previous implementation of search.
    #  list(filter(lambda x: ul <= x[0], sequence))[0]
    raise IndexError()


TypeOverrides = Mapping[str, Union[float, bool, None]]


def validate_overrides(data: TypeOverrides, config_model: Dict) -> None:
    """
    Check that override fields are valid.

    :param data: a dict of field name to value
    :param config_model: the configuration for the chosen model
    :raises ValueError: If a value is invalid
    """
    for key, value in data.items():
        field_config = config_model.get(key)
        is_quirk = key in config_model["quirks"] and key in VALID_QUIRKS

        if is_quirk:
            # If it's a quirk it must be a bool or None
            if value is not None and not isinstance(value, bool):
                raise ValueError(f"{value} is invalid for {key}")
        elif not field_config:
            # If it's not a quirk we must have a field config
            raise ValueError(f"Unknown field {key}")
        elif value is not None:
            # If value is not None it must be numeric and between min and max
            if isinstance(value, bool) or not isinstance(value, numbers.Number):
                raise ValueError(f"{value} is invalid for {key}")
            elif value < field_config["min"] or value > field_config["max"]:
                raise ValueError(f"{key} out of range with {value}")


def override(pipette_id: str, fields: TypeOverrides):
    """
    Override configuration for pipette. Validate then save.

    :param pipette_id: The pipette id
    :param fields: Dict of field name to override value
    """
    config_match = list_mutable_configs(pipette_id)
    whole_config, model = load_config_dict(pipette_id)
    validate_overrides(data=fields, config_model=config_match)
    save_overrides(pipette_id, fields, model)


def save_overrides(pipette_id: str, overrides: TypeOverrides, model: PipetteModel):
    """
    Save overrides for the pipette.

    :param pipette_id: The pipette id
    :param overrides: The incoming values
    :param model: The model of pipette
    :return: None
    """
    override_dir = config.CONFIG["pipette_config_overrides_dir"]
    model_configs = configs[model]
    model_configs_quirks = {key: True for key in model_configs["quirks"]}
    try:
        existing = load_overrides(pipette_id)
        # Add quirks setting for pipettes already with a pipette id file
        if "quirks" not in existing.keys():
            existing["quirks"] = model_configs_quirks
    except FileNotFoundError:
        existing = model_configs_quirks  # type: ignore

    for key, value in overrides.items():
        # If an existing override is saved as null from endpoint, remove from
        # overrides file
        if value is None:
            if existing.get(key):
                del existing[key]
        elif isinstance(value, bool):
            existing, model_configs = change_quirks(
                {key: value}, existing, model_configs
            )
        else:
            # type ignores are here because mypy needs typed dict accesses to
            # be string literals sadly enough
            model_config_value = model_configs[key]  # type: ignore
            if not model_config_value.get("default"):
                model_config_value["default"] = model_config_value["value"]
            model_config_value["value"] = value
            existing[key] = model_config_value
    assert model in config_models
    existing["model"] = model
    with (override_dir / f"{pipette_id}.json").open("w") as file:
        json.dump(existing, file)


def change_quirks(override_quirks, existing, model_configs):
    if not existing.get("quirks"):
        # ensure quirk key exists
        existing["quirks"] = override_quirks
    for quirk, setting in override_quirks.items():
        # setting values again if above case true, but
        # meant for use-cases where we may only be given an update
        # for one setting
        existing["quirks"][quirk] = setting
        if setting not in model_configs["quirks"]:
            model_configs["quirks"].append(quirk)
        elif not setting:
            model_configs["quirks"].remove(quirk)
    return existing, model_configs


def load_overrides(pipette_id: str) -> Dict[str, Any]:
    overrides = config.CONFIG["pipette_config_overrides_dir"]
    try:
        with (overrides / f"{pipette_id}.json").open() as fi:
            return json.load(fi)
    except json.JSONDecodeError as e:
        log.warning(f"pipette override for {pipette_id} is corrupt: {e}")
        (overrides / f"{pipette_id}.json").unlink()
        raise FileNotFoundError(str(overrides / f"{pipette_id}.json"))


def validate_quirks(quirks: List[str]):
    valid_quirks = []
    for quirk in quirks:
        if quirk in VALID_QUIRKS:
            valid_quirks.append(quirk)
        else:
            log.warning(f"{quirk} is not a valid quirk")
    return valid_quirks


def ensure_value(
    config: PipetteFusedSpec,
    name: Union[str, Tuple[str, ...]],
    mutable_config_list: List[str],
):
    """
    Pull value of config data from file. Shape can either be a dictionary with
    a value key -- indicating that it can be changed -- or another
    data structure such as an array.
    """
    if not isinstance(name, tuple):
        path: Tuple[str, ...] = (name,)
    else:
        path = name
    for element in path[:-1]:
        config = config[element]  # type: ignore

    value = config[path[-1]]  # type: ignore
    if path[-1] != "quirks" and path[-1] in mutable_config_list:
        value = value["value"]
    return value


def known_pipettes() -> Sequence[str]:
    """List pipette IDs for which we have known overrides"""
    return [
        fi.stem
        for fi in config.CONFIG["pipette_config_overrides_dir"].iterdir()
        if fi.is_file() and ".json" in fi.suffixes
    ]


def add_default(cfg):
    if isinstance(cfg, dict):
        if "value" in cfg.keys():
            cfg["default"] = cfg["value"]
        else:
            for top_level_key in cfg.keys():
                add_default(cfg[top_level_key])


def load_config_dict(pipette_id: str) -> Tuple["PipetteFusedSpec", "PipetteModel"]:
    """Give updated config with overrides for a pipette. This will add
    the default value for a mutable config before returning the modified
    config value.
    """
    override = load_overrides(pipette_id)
    model = override["model"]
    config = fuse_specs(model)

    if "quirks" not in override.keys():
        override["quirks"] = {key: True for key in config["quirks"]}
    else:
        # 20210210 AL - There have been bugs that allow settings invalid
        # quirks. Sanitize quirks by removing invalid ones that may have
        # been saved erroneously.
        override["quirks"] = {
            key: value
            for key, value in override["quirks"].items()
            if key in VALID_QUIRKS
        }

    for top_level_key in config.keys():
        if top_level_key != "quirks":
            add_default(config[top_level_key])  # type: ignore

    config.update(override)  # type: ignore

    return config, model


def list_mutable_configs(pipette_id: str) -> Dict[str, Any]:
    """
    Returns dict of mutable configs only.
    """
    cfg: Dict[str, Any] = {}

    try:
        config, model = load_config_dict(pipette_id)
    except FileNotFoundError:
        log.info(f"Pipette id {pipette_id} not found")
        return cfg

    for key in config:
        if key in MUTABLE_CONFIGS:
            cfg[key] = config[key]  # type: ignore
    return cfg



"""Base transport interfaces for communicating with a Protocol Engine."""
from abc import ABC, abstractmethod
from asyncio import AbstractEventLoop, run_coroutine_threadsafe

from ..protocol_engine import ProtocolEngine
from ..errors import ProtocolEngineError
from ..state import StateView
from ..commands import CommandRequest, CommandResult


class AbstractSyncTransport(ABC):
    """Interface describing a sync. ProtocolEngine state/command transport."""

    @property
    @abstractmethod
    def state(self) -> StateView:
        """Get a view of the ProtocolEngine's state."""
        ...

    @abstractmethod
    def execute_command(self, request: CommandRequest) -> CommandResult:
        """Execute a ProtocolEngine command, blocking until the command completes.

        Args:
            request: The ProtocolEngine command request

        Returns:
            The command's result data.

        Raises:
            ProtocolEngineError: if the command execution is not successful,
                the specific error that cause the command to fail is raised.
        """
        ...


class ChildThreadTransport(AbstractSyncTransport):
    """Concrete transport implementation using asyncio.run_coroutine_threadsafe."""

    def __init__(self, engine: ProtocolEngine, loop: AbstractEventLoop) -> None:
        """Initialize a ProtocolEngine transport for use in a child thread.

        This adapter allows a client to make blocking command calls on its
        thread to the asynchronous ProtocolEngine running with an event loop
        in a different thread.

        Args:
            engine: An instance of a ProtocolEngine to interact with hardware
                and other run procedures.
            loop: An event loop running in the thread where the hardware
                interaction should occur. The thread this loop is running
                in should be different than the thread in which the Python
                protocol is running.
        """
        self._engine = engine
        self._loop = loop

    @property
    def state(self) -> StateView:
        """Get a view of the Protocol Engine's state."""
        return self._engine.state_view

    def execute_command(self, request: CommandRequest) -> CommandResult:
        """Execute a command synchronously on the main thread."""
        command = run_coroutine_threadsafe(
            self._engine.add_and_execute_command(request=request),
            loop=self._loop,
        ).result()

        if command.error is not None:
            # TODO(mc, 2021-06-21): refactor when command.error is error details
            # rather than a string
            raise ProtocolEngineError(command.error)

        assert command.result is not None, f"Expected Command {command} to have result"

        return command.result



""" opentrons.calibration_storage.modify: functions for modifying
calibration storage

This module has functions that you can import to save robot or
labware calibration to its designated file location.
"""
import typing
from pathlib import Path

from dataclasses import asdict
from opentrons import config
from opentrons.types import Mount, Point

from opentrons.protocols.api_support.constants import OPENTRONS_NAMESPACE
from opentrons.util.helpers import utc_now

from . import file_operators as io, types as local_types, helpers, migration

if typing.TYPE_CHECKING:
    from .dev_types import (
        TipLengthCalibration,
        PipTipLengthCalibration,
        DeckCalibrationData,
        PipetteCalibrationData,
        CalibrationStatusDict,
    )
    from opentrons_shared_data.labware.dev_types import LabwareDefinition


def _add_to_index_offset_file(parent: str, slot: str, uri: str, lw_hash: str):
    """
    A helper method to create or add to an index file so that calibration
    files can be looked up by their hash to reveal the labware uri and
    parent information of a given file.

    :param parent: A labware object
    :param slot
    :param lw_hash: The labware hash of the calibration
    """
    offset = config.get_opentrons_path("labware_calibration_offsets_dir_v2")
    index_file = offset / "index.json"
    if index_file.exists():
        migration.check_index_version(index_file)
        blob = io.read_cal_file(str(index_file))
    else:
        blob = {}

    full_id = f"{lw_hash}{parent}"
    try:
        blob["data"][full_id]
    except KeyError:
        if parent:
            mod_dict = {"parent": parent, "fullParent": f"{slot}-{parent}"}
        else:
            mod_dict = {}
        new_index_data = {"uri": f"{uri}", "slot": full_id, "module": mod_dict}
        if blob.get("data"):
            blob["data"][full_id] = new_index_data
        else:
            blob["data"] = {full_id: new_index_data}
        blob["version"] = migration.MAX_VERSION
        io.save_to_file(index_file, blob)


def add_existing_labware_to_index_file(
    definition: "LabwareDefinition", parent: str = "", slot: str = ""
):
    labware_hash = helpers.hash_labware_def(definition)
    uri = helpers.uri_from_definition(definition)
    _add_to_index_offset_file(parent, slot, uri, labware_hash)


def save_labware_calibration(
    labware_path: local_types.StrPath,
    definition: "LabwareDefinition",
    delta: Point,
    slot: str = "",
    parent: str = "",
):
    """
    Function to be used whenever an updated delta is found for the first well
    of a given labware. If an offset file does not exist, create the file
    using labware id as the filename. If the file does exist, load it and
    modify the delta and the lastModified fields under the "default" key.

    :param labware_path: name of labware offset path
    :param definition: full definition of the labware
    :param delta: point you are saving
    :param slot: slot the labware calibration is associated with
    [not yet implemented so it will currently only be an empty string]
    :param parent: parent of the labware, either a slot or a module.
    """
    offset_path = config.get_opentrons_path("labware_calibration_offsets_dir_v2")
    labware_offset_path = offset_path / labware_path
    labware_hash = helpers.hash_labware_def(definition)
    uri = helpers.uri_from_definition(definition)
    _add_to_index_offset_file(parent, slot, uri, labware_hash)
    calibration_data = _helper_offset_data_format(str(labware_offset_path), delta)
    io.save_to_file(labware_offset_path, calibration_data)


def create_tip_length_data(
    definition: "LabwareDefinition",
    length: float,
    cal_status: local_types.CalibrationStatus = None,
) -> "PipTipLengthCalibration":
    """
    Function to correctly format tip length data.

    :param definition: full labware definition
    :param length: the tip length to save
    """
    labware_hash = helpers.hash_labware_def(definition)
    labware_uri = helpers.uri_from_definition(definition)
    if cal_status:
        status = cal_status
    else:
        status = local_types.CalibrationStatus()
    status_dict: "CalibrationStatusDict" = helpers.convert_to_dict(  # type: ignore[assignment]  # noqa: E501
        status
    )

    tip_length_data: "TipLengthCalibration" = {
        "tipLength": length,
        "lastModified": utc_now(),
        "source": local_types.SourceType.user,
        "status": status_dict,
        "uri": labware_uri,
    }

    if not definition.get("namespace") == OPENTRONS_NAMESPACE:
        _save_custom_tiprack_definition(labware_uri, definition)

    data = {labware_hash: tip_length_data}
    return data


def _save_custom_tiprack_definition(labware_uri: str, definition: "LabwareDefinition"):
    namespace, load_name, version = labware_uri.split("/")
    custom_tr_dir_path = config.get_custom_tiprack_def_path()
    custom_namespace_dir = custom_tr_dir_path / f"{namespace}/{load_name}"
    custom_namespace_dir.mkdir(parents=True, exist_ok=True)

    custom_tr_def_path = custom_namespace_dir / f"{version}.json"
    io.save_to_file(custom_tr_def_path, definition)


def _helper_offset_data_format(filepath: str, delta: Point) -> dict:
    if not Path(filepath).is_file():
        calibration_data = {
            "default": {
                "offset": [delta.x, delta.y, delta.z],
                "lastModified": utc_now(),
            }
        }
    else:
        calibration_data = io.read_cal_file(filepath)
        calibration_data["default"]["offset"] = [delta.x, delta.y, delta.z]
        calibration_data["default"]["lastModified"] = utc_now()
    return calibration_data


def _append_to_index_tip_length_file(pip_id: str, lw_hash: str):
    index_file = config.get_tip_length_cal_path() / "index.json"
    try:
        index_data = io.read_cal_file(str(index_file))
    except FileNotFoundError:
        index_data = {}

    if lw_hash not in index_data:
        index_data[lw_hash] = [pip_id]
    elif pip_id not in index_data[lw_hash]:
        index_data[lw_hash].append(pip_id)

    io.save_to_file(index_file, index_data)


def save_tip_length_calibration(pip_id: str, tip_length_cal: "PipTipLengthCalibration"):
    """
    Function used to save tip length calibration to file.

    :param pip_id: pipette id to associate with this tip length
    :param tip_length_cal: results of the data created using
           :meth:`create_tip_length_data`
    """
    tip_length_dir_path = config.get_tip_length_cal_path()
    tip_length_dir_path.mkdir(parents=True, exist_ok=True)
    pip_tip_length_path = tip_length_dir_path / f"{pip_id}.json"

    for lw_hash in tip_length_cal.keys():
        _append_to_index_tip_length_file(pip_id, lw_hash)

    try:
        tip_length_data = io.read_cal_file(str(pip_tip_length_path))
    except FileNotFoundError:
        tip_length_data = {}

    tip_length_data.update(tip_length_cal)

    io.save_to_file(pip_tip_length_path, tip_length_data)


def save_robot_deck_attitude(
    transform: local_types.AttitudeMatrix,
    pip_id: typing.Optional[str],
    lw_hash: typing.Optional[str],
    source: local_types.SourceType = None,
    cal_status: local_types.CalibrationStatus = None,
):
    robot_dir = config.get_opentrons_path("robot_calibration_dir")
    robot_dir.mkdir(parents=True, exist_ok=True)
    gantry_path = robot_dir / "deck_calibration.json"
    if cal_status:
        status = cal_status
    else:
        status = local_types.CalibrationStatus()
    status_dict: "CalibrationStatusDict" = helpers.convert_to_dict(  # type: ignore[assignment]  # noqa: E501
        status
    )

    gantry_dict: "DeckCalibrationData" = {
        "attitude": transform,
        "pipette_calibrated_with": pip_id,
        "last_modified": utc_now(),
        "tiprack": lw_hash,
        "source": source or local_types.SourceType.user,
        "status": status_dict,
    }
    io.save_to_file(gantry_path, gantry_dict)


def _add_to_pipette_offset_index_file(pip_id: str, mount: Mount):
    index_file = config.get_opentrons_path("pipette_calibration_dir") / "index.json"
    try:
        index_data = index_data = io.read_cal_file(str(index_file))
    except FileNotFoundError:
        index_data = {}

    mount_key = mount.name.lower()
    if mount_key not in index_data:
        index_data[mount_key] = [pip_id]
    elif pip_id not in index_data[mount_key]:
        index_data[mount_key].append(pip_id)

    io.save_to_file(index_file, index_data)


def save_pipette_calibration(
    offset: Point,
    pip_id: str,
    mount: Mount,
    tiprack_hash: str,
    tiprack_uri: str,
    cal_status: local_types.CalibrationStatus = None,
):
    pip_dir = config.get_opentrons_path("pipette_calibration_dir") / mount.name.lower()
    pip_dir.mkdir(parents=True, exist_ok=True)
    if cal_status:
        status = cal_status
    else:
        status = local_types.CalibrationStatus()
    status_dict: "CalibrationStatusDict" = helpers.convert_to_dict(  # type: ignore[assignment]  # noqa: E501
        status
    )
    offset_path = pip_dir / f"{pip_id}.json"
    offset_dict: "PipetteCalibrationData" = {
        "offset": [offset.x, offset.y, offset.z],
        "tiprack": tiprack_hash,
        "uri": tiprack_uri,
        "last_modified": utc_now(),
        "source": local_types.SourceType.user,
        "status": status_dict,
    }
    io.save_to_file(offset_path, offset_dict)
    _add_to_pipette_offset_index_file(pip_id, mount)


@typing.overload
def mark_bad(
    calibration: local_types.DeckCalibration, source_marked_bad: local_types.SourceType
) -> local_types.DeckCalibration:
    ...


@typing.overload
def mark_bad(
    calibration: local_types.PipetteOffsetCalibration,
    source_marked_bad: local_types.SourceType,
) -> local_types.PipetteOffsetCalibration:
    ...


@typing.overload
def mark_bad(
    calibration: local_types.TipLengthCalibration,
    source_marked_bad: local_types.SourceType,
) -> local_types.TipLengthCalibration:
    ...


def mark_bad(calibration, source_marked_bad: local_types.SourceType):
    caldict = asdict(calibration)
    # remove current status key
    del caldict["status"]
    status = local_types.CalibrationStatus(
        markedBad=True, source=source_marked_bad, markedAt=utc_now()
    )
    return type(calibration)(**caldict, status=status)



"""Command queue execution worker module."""
import asyncio
from logging import getLogger
from typing import Optional

from ..state import StateStore
from ..errors import ProtocolEngineStoppedError
from .command_executor import CommandExecutor


log = getLogger(__name__)


class QueueWorker:
    """Handle and track execution of commands queued in ProtocolEngine state."""

    def __init__(
        self,
        state_store: StateStore,
        command_executor: CommandExecutor,
    ) -> None:
        """Initialize the queue worker's dependencies and state.

        Arguments:
            state_store: The source of truth for protocol state, including
                all queued commands.
            command_executor: Interface used to execute and update commands.
        """
        self._state_store: StateStore = state_store
        self._command_executor: CommandExecutor = command_executor
        self._worker_task: Optional["asyncio.Task[None]"] = None

    def start(self) -> None:
        """Start processing jobs.

        This method will no-op if the worker is already running.
        """
        if self._worker_task is None:
            self._worker_task = asyncio.create_task(self._run_commands())

    def cancel(self) -> None:
        """Cancel any in-progress commands.

        This method is synchronous to allow synchronous callers to
        cancel the ongoing background task in situations where it
        needs to happen immediately.

        You should call `join` after calling `cancel` to clean up and
        propagate errors.
        """
        if self._worker_task:
            self._worker_task.cancel()

    async def join(self) -> None:
        """Wait for the worker to finish, propagating any errors."""
        worker_task = self._worker_task

        if worker_task:
            self._worker_task = None

            try:
                await worker_task
            except (ProtocolEngineStoppedError, asyncio.CancelledError):
                pass
            except Exception as e:
                log.error("Unhandled exception in QueueWorker job", exc_info=e)
                raise e

    async def _run_commands(self) -> None:
        while not self._state_store.commands.get_stop_requested():
            command_id = await self._state_store.wait_for(
                condition=self._state_store.commands.get_next_queued
            )

            await self._command_executor.execute(command_id=command_id)



"""Public protocol engine value types and models."""
from enum import Enum
from dataclasses import dataclass
from pydantic import BaseModel
from typing import Union, Tuple

from opentrons.types import MountType, DeckSlotName


class EngineStatus(str, Enum):
    """Current execution status of a ProtocolEngine."""

    READY_TO_RUN = "ready-to-run"
    RUNNING = "running"
    PAUSE_REQUESTED = "pause-requested"
    PAUSED = "paused"
    STOP_REQUESTED = "stop-requested"
    STOPPED = "stopped"
    FAILED = "failed"
    SUCCEEDED = "succeeded"


class DeckSlotLocation(BaseModel):
    """Location for labware placed in a single slot."""

    slot: DeckSlotName


LabwareLocation = Union[DeckSlotLocation]
"""Union of all legal labware locations."""


class WellOrigin(str, Enum):
    """Origin of WellLocation offset."""

    TOP = "top"
    BOTTOM = "bottom"


class WellLocation(BaseModel):
    """A relative location in reference to a well's location."""

    origin: WellOrigin = WellOrigin.TOP
    offset: Tuple[float, float, float] = (0, 0, 0)


@dataclass(frozen=True)
class Dimensions:
    """Dimensions of an object in deck-space."""

    x: float
    y: float
    z: float


class CalibrationOffset(BaseModel):
    """Calibration offset from nomimal to actual position."""

    x: float
    y: float
    z: float


# TODO(mc, 2021-04-16): reconcile with opentrons_shared_data
# shared-data/python/opentrons_shared_data/pipette/dev_types.py
class PipetteName(str, Enum):
    """Pipette load name values."""

    P10_SINGLE = "p10_single"
    P10_MULTI = "p10_multi"
    P20_SINGLE_GEN2 = "p20_single_gen2"
    P20_MULTI_GEN2 = "p20_multi_gen2"
    P50_SINGLE = "p50_single"
    P50_MULTI = "p50_multi"
    P300_SINGLE = "p300_single"
    P300_MULTI = "p300_multi"
    P300_SINGLE_GEN2 = "p300_single_gen2"
    P300_MULTI_GEN2 = "p300_multi_gen2"
    P1000_SINGLE = "p1000_single"
    P1000_SINGLE_GEN2 = "p1000_single_gen2"


class LoadedPipette(BaseModel):
    """A pipette that has been loaded."""

    id: str
    pipetteName: PipetteName
    mount: MountType


class LoadedLabware(BaseModel):
    """A labware that has been loaded."""

    id: str
    loadName: str
    definitionUri: str
    location: LabwareLocation



from __future__ import annotations
import asyncio
from asyncio import Queue
from contextlib import contextmanager
import logging
from typing import (
    Any,
    Callable,
    Dict,
    Sequence,
    overload,
    Generic,
    TypeVar,
    cast,
    TYPE_CHECKING,
)
from typing_extensions import Literal

from opentrons.commands import types

if TYPE_CHECKING:
    from opentrons.api.dev_types import Message as SessionStateMessage
    from opentrons.api.calibration import Message as CalibrationStateMessage


MODULE_LOG = logging.getLogger(__name__)


UntypedMessage = Dict[str, Any]


_HandledMessages = TypeVar("_HandledMessages")


class Notifications(Generic[_HandledMessages]):
    def __init__(self, topics: Sequence[str], broker: Broker, loop=None):
        self.loop = loop or asyncio.get_event_loop()
        self.queue: Queue[UntypedMessage] = Queue(loop=self.loop)
        self.snoozed = False
        self._unsubscribe = [
            broker.subscribe(topic, self.on_notify) for topic in topics
        ]

    @contextmanager
    def snooze(self):
        self.snoozed = True
        try:
            yield
        finally:
            self.snoozed = False

    def on_notify(self, message: UntypedMessage):
        if self.snoozed:
            return
        self.queue.put_nowait(message)

    async def __anext__(self) -> _HandledMessages:
        msg = await self.queue.get()
        return cast(_HandledMessages, msg)

    def __aiter__(self):
        return self


class Broker:
    def __init__(self):
        self.subscriptions = {}
        self.logger = MODULE_LOG

    @overload
    def subscribe(
        self, topic: Literal["command"], handler: Callable[[types.CommandMessage], None]
    ) -> Callable[[], None]:
        ...

    @overload
    def subscribe(
        self,
        topic: Literal["calibration"],
        handler: Callable[[CalibrationStateMessage], None],
    ) -> Callable[[], None]:
        ...

    @overload
    def subscribe(
        self, topic: Literal["session"], handler: Callable[[SessionStateMessage], None]
    ) -> Callable[[], None]:
        ...

    @overload
    def subscribe(
        self, topic: str, handler: Callable[[UntypedMessage], None]
    ) -> Callable[[], None]:
        ...

    def subscribe(self, topic, handler):
        if handler in self.subscriptions.setdefault(topic, []):
            return
        self.subscriptions[topic].append(handler)

        def unsubscribe():
            self.subscriptions[topic].remove(handler)

        return unsubscribe

    @overload
    def publish(self, topic: Literal["command"], message: types.CommandMessage) -> None:
        ...

    @overload
    def publish(
        self, topic: Literal["calibration"], message: CalibrationStateMessage
    ) -> None:
        ...

    @overload
    def publish(self, topic: Literal["session"], message: SessionStateMessage) -> None:
        ...

    @overload
    def publish(self, topic: str, message: UntypedMessage) -> None:
        ...

    def publish(self, topic, message):
        [handler(message) for handler in self.subscriptions.get(topic, [])]

    def set_logger(self, logger):
        self.logger = logger



"""Algorithm types.

Any type definitions required for the algorithms
module should be put in this file.
"""

from typing import TypeVar, Generic, List
from dataclasses import dataclass


VertexName = TypeVar("VertexName")


@dataclass(frozen=True)
class GenericNode(Generic[VertexName]):
    """Generic graph node dataclass.

    A dataclass to hold information about a
    graph node. Information should not be
    mutated about the node.
    """

    name: VertexName
    sub_names: List[VertexName]

    def __hash__(self) -> int:
        """Hash function.

        To have a unique set of nodes, they must
        all have a unique hash. Lists are not
        hashable which is why we need to unpack
        the list here.
        """
        return hash((self.name, *self.sub_names))


VertexLike = TypeVar("VertexLike", bound=GenericNode)



from dataclasses import dataclass
from typing import Dict, cast

from opentrons.config.types import AxisDict

AxisSettingType = Dict[str, float]


@dataclass
class AxisCurrentSettings:
    now: AxisSettingType
    saved: AxisSettingType

    def __init__(self, val: AxisDict) -> None:
        self.now = cast(AxisSettingType, val.copy())
        self.saved = cast(AxisSettingType, val.copy())



from .driver import TempDeckDriver
from .abstract import AbstractTempDeckDriver
from .simulator import SimulatingDriver


__all__ = [
    "TempDeckDriver",
    "AbstractTempDeckDriver",
    "SimulatingDriver",
]



"""QueueWorker and dependency factory."""
from opentrons.hardware_control.api import API as HardwareAPI

from ..state import StateStore
from .equipment import EquipmentHandler
from .movement import MovementHandler
from .pipetting import PipettingHandler
from .run_control import RunControlHandler
from .command_executor import CommandExecutor
from .queue_worker import QueueWorker


def create_queue_worker(
    hardware_api: HardwareAPI,
    state_store: StateStore,
) -> QueueWorker:
    """Create a ready-to-use QueueWorker instance.

    Arguments:
        hardware_api: Hardware control API to pass down to dependencies.
        state_store: The ProtocolEngine StateStore instance to pass down
            to dependencies.
    """
    equipment_handler = EquipmentHandler(
        hardware_api=hardware_api,
        state_store=state_store,
    )

    movement_handler = MovementHandler(
        hardware_api=hardware_api,
        state_store=state_store,
    )

    pipetting_handler = PipettingHandler(
        hardware_api=hardware_api,
        state_store=state_store,
        movement_handler=movement_handler,
    )

    run_control_handler = RunControlHandler(state_store=state_store)

    command_executor = CommandExecutor(
        state_store=state_store,
        equipment=equipment_handler,
        movement=movement_handler,
        pipetting=pipetting_handler,
        run_control=run_control_handler,
    )

    return QueueWorker(
        state_store=state_store,
        command_executor=command_executor,
    )



# noqa: D100

from typing import List, Optional, Sequence, Union

from opentrons.protocol_engine.clients import SyncClient as ProtocolEngineClient

from .pipette_context import PipetteContext
from .instrument_context import InstrumentContext
from .labware import Labware
from .types import DeckSlotName, DeckSlotLocation, DeprecatedMount, Mount, PipetteName
from . import errors


class ProtocolContext:  # noqa: D101
    def __init__(self, engine_client: ProtocolEngineClient) -> None:
        """Initialize a ProtocolContext API provider.

        You do not need to initialize the ProtocolContext yourself; the system
        will create one and pass it to your protocol's `run` method.

        Args:
            engine_client: A ProtocolEngine client to issue protocol commands.
        """
        self._engine_client = engine_client

    def load_pipette(  # noqa: D102
        self,
        pipette_name: str,
        mount: str,
        tip_racks: Sequence[Labware] = (),
        replace: bool = False,
    ) -> PipetteContext:
        if pipette_name not in list(PipetteName):
            raise errors.InvalidPipetteNameError(pipette_name)

        if mount not in list(Mount):
            raise errors.InvalidMountError(mount)

        if len(tip_racks) > 0:
            # TODO(mc, 2021-04-16): figure out what to do with `tip_racks`
            raise NotImplementedError()

        if replace:
            # TODO(mc, 2021-04-16): figure out what to do with `replace`
            raise NotImplementedError()

        result = self._engine_client.load_pipette(
            pipette_name=PipetteName(pipette_name),
            mount=Mount(mount),
        )

        return PipetteContext(
            engine_client=self._engine_client,
            pipette_id=result.pipetteId,
        )

    def load_instrument(
        self,
        instrument_name: str,
        mount: Union[DeprecatedMount, str],
        tip_racks: Optional[List[Labware]] = None,
        replace: bool = False,
    ) -> InstrumentContext:
        """Load a pipette into the protocol.

        .. deprecated:: Protocol API v3.0
            Use :py:meth:`load_pipette` instead.
        """
        return self.load_pipette(
            pipette_name=instrument_name,
            mount=(mount if isinstance(mount, str) else str(mount).lower()),
            tip_racks=(tip_racks if tip_racks is not None else ()),
            replace=replace,
        )

    def load_labware(  # noqa: D102
        self,
        load_name: str,
        location: Union[int, str],
        label: Optional[str] = None,
        namespace: Optional[str] = None,
        version: Optional[int] = None,
    ) -> Labware:
        if label is not None:
            raise NotImplementedError("Labware labeling not yet implemented.")

        result = self._engine_client.load_labware(
            load_name=load_name,
            location=DeckSlotLocation(slot=DeckSlotName.from_primitive(location)),
            # TODO(mc, 2021-04-22): make sure this default is compatible with using
            # namespace=None to load custom labware in PAPIv3
            namespace=namespace or "opentrons",
            version=version or 1,
        )

        return Labware(engine_client=self._engine_client, labware_id=result.labwareId)

    def pause(self, msg: Optional[str] = None) -> None:
        """Pause execution of the protocol until resumed by the user.

        This method will block until the protocol is resumed.

        Arguments:
            msg: An optional message to attach to the pause command. This message
                will be available to the app or client controlling the protocol.
        """
        self._engine_client.pause(message=msg)

    # todo(mm, 2021-04-09): Add all other public methods from the APIv2
    # ProtocolContext.



from typing import List, Set
from typing_extensions import Protocol

from opentrons.hardware_control.modules.types import ModuleAtPort
from opentrons.hardware_control.types import BoardRevision

from .types import USBPort


class USBDriverInterface(Protocol):
    @staticmethod
    def read_bus() -> List[str]:
        ...

    @staticmethod
    def read_symlink(virtual_port: str) -> str:
        ...

    @property
    def board_revision(self) -> BoardRevision:
        ...

    @property
    def usb_dev(self) -> List[USBPort]:
        ...

    @usb_dev.setter
    def usb_dev(self, ports: List[USBPort]) -> None:
        ...

    @property
    def sorted_ports(self) -> Set:
        ...

    @sorted_ports.setter
    def sorted_ports(self, sorted: Set) -> None:
        ...

    def read_usb_bus(self) -> List[USBPort]:
        ...

    def find_port(self, device_path: str) -> USBPort:
        ...

    def sort_ports(self) -> None:
        ...

    def match_virtual_ports(
        self, virtual_port: List[ModuleAtPort]
    ) -> List[ModuleAtPort]:
        ...



"""An emulation of the opentrons temperature module.

The purpose is to provide a fake backend that responds to GCODE commands.
"""

import logging
from typing import Optional

from opentrons.drivers.temp_deck.driver import GCODE
from opentrons.hardware_control.emulation import util
from opentrons.hardware_control.emulation.parser import Parser, Command

from .abstract_emulator import AbstractEmulator
from .simulations import Temperature


logger = logging.getLogger(__name__)


SERIAL = "temperature_emulator"
MODEL = "temp_deck_v20"
VERSION = "v2.0.1"


class TempDeckEmulator(AbstractEmulator):
    """TempDeck emulator"""

    def __init__(self, parser: Parser) -> None:
        self.reset()
        self._parser = parser

    def handle(self, line: str) -> Optional[str]:
        """Handle a line"""
        results = (self._handle(c) for c in self._parser.parse(line))
        joined = " ".join(r for r in results if r)
        return None if not joined else joined

    def reset(self):
        self._temperature = Temperature(per_tick=0.25, current=0.0)

    def _handle(self, command: Command) -> Optional[str]:
        """Handle a command."""
        logger.info(f"Got command {command}")
        if command.gcode == GCODE.GET_TEMP:
            res = (
                f"T:{util.OptionalValue(self._temperature.target)} "
                f"C:{self._temperature.current}"
            )
            self._temperature.tick()
            return res
        elif command.gcode == GCODE.SET_TEMP:
            temperature = command.params["S"]
            assert isinstance(
                temperature, float
            ), f"invalid temperature '{temperature}'"
            self._temperature.set_target(temperature)
        elif command.gcode == GCODE.DISENGAGE:
            self._temperature.deactivate(util.TEMPERATURE_ROOM)
        elif command.gcode == GCODE.DEVICE_INFO:
            return f"serial:{SERIAL} model:{MODEL} version:{VERSION}"
        elif command.gcode == GCODE.PROGRAMMING_MODE:
            pass
        return None



"""Map command request and model types to other values."""
from datetime import datetime
from typing import Any

from .command import CommandStatus
from .command_unions import Command, CommandRequest


class CommandMapper:
    """Static methods to map and modify command models."""

    @staticmethod
    def map_request_to_command(
        request: CommandRequest,
        command_id: str,
        created_at: datetime,
    ) -> Command:
        """Map a CommandRequest instance to a full command."""
        # TODO(mc, 2021-06-22): mypy has trouble with this automatic
        # request > command mapping, figure out how to type precisely
        # (or wait for a future mypy version that can figure it out).
        # For now, unit tests cover mapping every request type
        return request._CommandCls(
            id=command_id,
            createdAt=created_at,
            status=CommandStatus.QUEUED,
            data=request.data,  # type: ignore[arg-type]
        )

    @staticmethod
    def update_command(command: Command, **update: Any) -> Command:
        """Map a Command to a new Command instance with updated field values.

        This is a thin wrapper around `BaseModel.copy`, but it ensures that
        all updates to commands happen immutably.
        """
        return command.copy(update=update)



""" opentrons.execute: functions and entrypoint for running protocols

This module has functions that can be imported to provide protocol
contexts for running protocols during interactive sessions like Jupyter or just
regular python shells. It also provides a console entrypoint for running a
protocol from the command line.
"""
import atexit
import argparse
import logging
import os
import sys
from typing import Any, Callable, Dict, List, Optional, TextIO, Union, TYPE_CHECKING

import opentrons
from opentrons import protocol_api, __version__
from opentrons.config import IS_ROBOT, JUPYTER_NOTEBOOK_LABWARE_DIR
from opentrons.protocol_api import MAX_SUPPORTED_VERSION
from opentrons.protocols.execution import execute as execute_apiv2
from opentrons.protocols.context.protocol_api.protocol_context import (
    ProtocolContextImplementation,
)
from opentrons.commands import types as command_types
from opentrons.protocols.parse import parse, version_from_string
from opentrons.protocols.types import PythonProtocol
from opentrons.protocols.api_support.types import APIVersion
from opentrons.hardware_control import API, ThreadManager
from .util.entrypoint_util import labware_from_paths, datafiles_from_paths

if TYPE_CHECKING:
    from opentrons_shared_data.labware.dev_types import LabwareDefinition

_THREAD_MANAGED_HW: Optional[ThreadManager] = None
#: The background global cache that all protocol contexts created by
#: :py:meth:`get_protocol_api` will share


def get_protocol_api(
    version: Union[str, APIVersion],
    bundled_labware: Dict[str, "LabwareDefinition"] = None,
    bundled_data: Dict[str, bytes] = None,
    extra_labware: Dict[str, "LabwareDefinition"] = None,
) -> protocol_api.ProtocolContext:
    """
    Build and return a :py:class:`ProtocolContext` connected to the robot.

    This can be used to run protocols from interactive Python sessions
    such as Jupyter or an interpreter on the command line:

    .. code-block:: python

        >>> from opentrons.execute import get_protocol_api
        >>> protocol = get_protocol_api('2.0')
        >>> instr = protocol.load_instrument('p300_single', 'right')
        >>> instr.home()

    If ``extra_labware`` is not specified, any labware definitions saved in
    the ``labware`` directory of the Jupyter notebook directory will be
    available.

    When this function is called, modules and instruments will be recached.

    :param version: The API version to use. This must be lower than
                    :py:attr:`opentrons.protocol_api.MAX_SUPPORTED_VERSION`.
                    It may be specified either as a string (``'2.0'``) or
                    as a :py:class:`.protocols.types.APIVersion`
                    (``APIVersion(2, 0)``).
    :param bundled_labware: If specified, a mapping from labware names to
                            labware definitions for labware to consider in the
                            protocol. Note that if you specify this, _only_
                            labware in this argument will be allowed in the
                            protocol. This is preparation for a beta feature
                            and is best not used.
    :param bundled_data: If specified, a mapping from filenames to contents
                         for data to be available in the protocol from
                         :py:attr:`.ProtocolContext.bundled_data`.
    :param extra_labware: If specified, a mapping from labware names to
                          labware definitions for labware to consider in the
                          protocol in addition to those stored on the robot.
                          If this is an empty dict, and this function is called
                          on a robot, it will look in the 'labware'
                          subdirectory of the Jupyter data directory for
                          custom labware.
    :returns opentrons.protocol_api.ProtocolContext: The protocol context.
    """
    global _THREAD_MANAGED_HW
    if not _THREAD_MANAGED_HW:
        # Build a hardware controller in a worker thread, which is necessary
        # because ipython runs its notebook in asyncio but the notebook
        # is at script/repl scope not function scope and is synchronous so
        # you can't control the loop from inside. If we update to
        # IPython 7 we can avoid this, but for now we can't
        _THREAD_MANAGED_HW = ThreadManager(API.build_hardware_controller)
    if isinstance(version, str):
        checked_version = version_from_string(version)
    elif not isinstance(version, APIVersion):
        raise TypeError("version must be either a string or an APIVersion")
    else:
        checked_version = version

    if (
        extra_labware is None
        and IS_ROBOT
        and JUPYTER_NOTEBOOK_LABWARE_DIR.is_dir()  # type: ignore[union-attr]
    ):
        extra_labware = labware_from_paths([str(JUPYTER_NOTEBOOK_LABWARE_DIR)])

    context_imp = ProtocolContextImplementation(
        hardware=_THREAD_MANAGED_HW,
        bundled_labware=bundled_labware,
        bundled_data=bundled_data,
        extra_labware=extra_labware,
        api_version=checked_version,
    )

    context = protocol_api.ProtocolContext(
        implementation=context_imp, api_version=checked_version
    )
    context_imp.get_hardware().hardware.cache_instruments()
    return context


def get_arguments(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    """Get the argument parser for this module

    Useful if you want to use this module as a component of another CLI program
    and want to add its arguments.

    :param parser: A parser to add arguments to.
    :returns argparse.ArgumentParser: The parser with arguments added.
    """
    parser.add_argument(
        "-l",
        "--log-level",
        choices=["debug", "info", "warning", "error", "none"],
        default="warning",
        help="Specify the level filter for logs to show on the command line. "
        "The logs stored in journald or local log files are unaffected by "
        "this option and should be configured in the config file. If "
        "'none', do not show logs",
    )
    parser.add_argument(
        "-L",
        "--custom-labware-path",
        action="append",
        default=[os.getcwd()],
        help="Specify directories to search for custom labware definitions. "
        "You can specify this argument multiple times. Once you specify "
        "a directory in this way, labware definitions in that directory "
        "will become available in ProtocolContext.load_labware(). "
        "Only directories specified directly by "
        "this argument are searched, not their children. JSON files that "
        "do not define labware will be ignored with a message. "
        "By default, the current directory (the one in which you are "
        "invoking this program) will be searched for labware.",
    )
    parser.add_argument(
        "-D",
        "--custom-data-path",
        action="append",
        nargs="?",
        const=".",
        default=[],
        help="Specify directories to search for custom data files. "
        "You can specify this argument multiple times. Once you specify "
        "a directory in this way, files located in the specified "
        "directory will be available in ProtocolContext.bundled_data. "
        "Note that bundle execution will still only allow data files in "
        "the bundle. If you specify this without a path, it will "
        "add the current path implicitly. If you do not specify this "
        "argument at all, no data files will be added. Any file in the "
        "specified paths will be loaded into memory and included in the "
        "bundle if --bundle is passed, so be careful that any directory "
        "you specify has only the files you want. It is usually a "
        "better idea to use -d so no files are accidentally included. "
        "Also note that data files are made available as their name, not "
        "their full path, so name them uniquely.",
    )
    parser.add_argument(
        "-d",
        "--custom-data-file",
        action="append",
        default=[],
        help="Specify data files to be made available in "
        "ProtocolContext.bundled_data (and possibly bundled if --bundle "
        "is passed). Can be specified multiple times with different "
        "files. It is usually a better idea to use this than -D because "
        "there is less possibility of accidentally including something.",
    )
    parser.add_argument(
        "protocol",
        metavar="PROTOCOL",
        type=argparse.FileType("rb"),
        help="The protocol file to execute. If you pass '-', you can pipe "
        "the protocol via stdin; this could be useful if you want to use this "
        "utility as part of an automated workflow.",
    )
    return parser


def execute(
    protocol_file: TextIO,
    protocol_name: str,
    propagate_logs: bool = False,
    log_level: str = "warning",
    emit_runlog: Callable[[Dict[str, Any]], None] = None,
    custom_labware_paths: List[str] = None,
    custom_data_paths: List[str] = None,
):
    """
    Run the protocol itself.

    This is a one-stop function to run a protocol, whether python or json,
    no matter the api verson, from external (i.e. not bound up in other
    internal server infrastructure) sources.

    To run an opentrons protocol from other places, pass in a file like
    object as protocol_file; this function either returns (if the run has no
    problems) or raises an exception.

    To call from the command line use either the autogenerated entrypoint
    ``opentrons_execute`` or ``python -m opentrons.execute``.

    If the protocol is using Opentrons Protocol API V1, it does not need to
    explicitly call :py:meth:`.Robot.connect`
    or :py:meth:`.Robot.discover_modules`, or
    :py:meth:`.Robot.cache_instrument_models`.

    :param file-like protocol_file: The protocol file to execute
    :param str protocol_name: The name of the protocol file. This is required
                              internally, but it may not be a thing we can get
                              from the protocol_file argument.
    :param propagate_logs: Whether this function should allow logs from the
                           Opentrons stack to propagate up to the root handler.
                           This can be useful if you're integrating this
                           function in a larger application, but most logs that
                           occur during protocol simulation are best associated
                           with the actions in the protocol that cause them.
                           Default: ``False``
    :type propagate_logs: bool
    :param log_level: The level of logs to emit on the command line.. Default:
                      'warning'
    :type log_level: 'debug', 'info', 'warning', or 'error'
    :param emit_runlog: A callback for printing the runlog. If specified, this
                        will be called whenever a command adds an entry to the
                        runlog, which can be used for display and progress
                        estimation. If specified, the callback should take a
                        single argument (the name doesn't matter) which will
                        be a dictionary (see below). Default: ``None``
    :param custom_labware_paths: A list of directories to search for custom
                                 labware, or None. Ignored if the apiv2 feature
                                 flag is not set. Loads valid labware from
                                 these paths and makes them available to the
                                 protocol context.
    :param custom_data_paths: A list of directories or files to load custom
                              data files from. Ignored if the apiv2 feature
                              flag if not set. Entries may be either files or
                              directories. Specified files and the
                              non-recursive contents of specified directories
                              are presented by the protocol context in
                              :py:attr:`.ProtocolContext.bundled_data`.

    The format of the runlog entries is as follows:

    .. code-block:: python

        {
            'name': command_name,
            'payload': {
                 'text': string_command_text,
                  # The rest of this struct is command-dependent; see
                  # opentrons.commands.commands. Its keys match format
                  # keys in 'text', so that
                  # entry['payload']['text'].format(**entry['payload'])
                  # will produce a string with information filled in
             }
        }


    """
    stack_logger = logging.getLogger("opentrons")
    stack_logger.propagate = propagate_logs
    stack_logger.setLevel(getattr(logging, log_level.upper(), logging.WARNING))
    contents = protocol_file.read()
    if custom_labware_paths:
        extra_labware = labware_from_paths(custom_labware_paths)
    else:
        extra_labware = {}
    if custom_data_paths:
        extra_data = datafiles_from_paths(custom_data_paths)
    else:
        extra_data = {}
    protocol = parse(
        contents, protocol_name, extra_labware=extra_labware, extra_data=extra_data
    )
    if getattr(protocol, "api_level", APIVersion(2, 0)) < APIVersion(2, 0):
        opentrons.robot.connect()
        opentrons.robot.cache_instrument_models()
        opentrons.robot.discover_modules()
        opentrons.robot.home()
        if emit_runlog:
            opentrons.robot.broker.subscribe(command_types.COMMAND, emit_runlog)
        assert isinstance(
            protocol, PythonProtocol
        ), "Internal error: Only Python protocols may be executed in v1"
        exec(protocol.contents, {})
    else:
        bundled_data = getattr(protocol, "bundled_data", {})
        bundled_data.update(extra_data)
        gpa_extras = getattr(protocol, "extra_labware", None) or None
        context = get_protocol_api(
            getattr(protocol, "api_level", MAX_SUPPORTED_VERSION),
            bundled_labware=getattr(protocol, "bundled_labware", None),
            bundled_data=bundled_data,
            extra_labware=gpa_extras,
        )
        if emit_runlog:
            broker = context.broker
            broker.subscribe(command_types.COMMAND, emit_runlog)
        context.home()
        try:
            execute_apiv2.run_protocol(protocol, context)
        finally:
            context.cleanup()


def make_runlog_cb():
    level = 0
    last_dollar = None

    def _print_runlog(command: Dict[str, Any]):
        nonlocal level
        nonlocal last_dollar

        if last_dollar == command["$"]:
            if command["$"] == "before":
                level += 1
            else:
                level -= 1
        last_dollar = command["$"]
        if command["$"] == "before":
            print(
                " ".join(
                    [
                        "\t" * level,
                        command["payload"].get("text", "").format(**command["payload"]),
                    ]
                )
            )

    return _print_runlog


def main() -> int:
    """Handler for command line invocation to run a protocol.

    :param argv: The arguments the program was invoked with; this is usually
                 :py:attr:`sys.argv` but if you want to override that you can.
    :returns int: A success or failure value suitable for use as a shell
                  return code passed to :py:meth:`sys.exit` (0 means success,
                  anything else is a kind of failure).
    """
    parser = argparse.ArgumentParser(
        prog="opentrons_execute", description="Run an OT-2 protocol"
    )
    parser = get_arguments(parser)
    # don't want to add this in get_arguments because if somebody upstream is
    # using that parser they probably want their own version
    parser.add_argument("-v", "--version", action="version", version=__version__)
    parser.add_argument(
        "-n",
        "--no-print-runlog",
        action="store_true",
        help="Do not print the commands as they are executed",
    )
    args = parser.parse_args()
    printer = None if args.no_print_runlog else make_runlog_cb()
    if args.log_level != "none":
        stack_logger = logging.getLogger("opentrons")
        stack_logger.addHandler(logging.StreamHandler(sys.stdout))
        log_level = args.log_level
    else:
        log_level = "warning"
    # Try to migrate containers from database to v2 format
    execute(args.protocol, args.protocol.name, log_level=log_level, emit_runlog=printer)
    return 0


@atexit.register
def _clear_cached_hardware_controller():
    global _THREAD_MANAGED_HW
    if _THREAD_MANAGED_HW:
        _THREAD_MANAGED_HW.clean_up()
        _THREAD_MANAGED_HW = None


if __name__ == "__main__":
    sys.exit(main())



from abc import ABC, abstractmethod
from typing import Dict


class AbstractMagDeckDriver(ABC):
    @abstractmethod
    async def connect(self) -> None:
        ...

    @abstractmethod
    async def disconnect(self) -> None:
        ...

    @abstractmethod
    async def is_connected(self) -> bool:
        pass

    @abstractmethod
    async def home(self) -> None:
        ...

    @abstractmethod
    async def probe_plate(self) -> None:
        ...

    @abstractmethod
    async def get_plate_height(self) -> float:
        ...

    @abstractmethod
    async def get_mag_position(self) -> float:
        ...

    @abstractmethod
    async def move(self, position_mm: float) -> None:
        ...

    @abstractmethod
    async def get_device_info(self) -> Dict[str, str]:
        ...

    @abstractmethod
    async def enter_programming_mode(self) -> None:
        ...



"""Equipment command side-effect logic."""
from dataclasses import dataclass
from typing import Tuple, Optional

from opentrons.calibration_storage.helpers import uri_from_details
from opentrons.protocols.models import LabwareDefinition
from opentrons.types import MountType
from opentrons.hardware_control.api import API as HardwareAPI

from ..errors import FailedToLoadPipetteError, LabwareDefinitionDoesNotExistError
from ..resources import LabwareDataProvider, ModelUtils
from ..state import StateStore
from ..types import LabwareLocation, PipetteName


@dataclass(frozen=True)
class LoadedLabwareData:
    """The result of a load labware procedure."""

    labware_id: str
    definition: LabwareDefinition
    calibration: Tuple[float, float, float]


@dataclass(frozen=True)
class LoadedPipetteData:
    """The result of a load pipette procedure."""

    pipette_id: str


class EquipmentHandler:
    """Implementation logic for labware, pipette, and module loading."""

    _hardware_api: HardwareAPI
    _state_store: StateStore
    _labware_data_provider: LabwareDataProvider
    _model_utils: ModelUtils

    def __init__(
        self,
        hardware_api: HardwareAPI,
        state_store: StateStore,
        labware_data_provider: Optional[LabwareDataProvider] = None,
        model_utils: Optional[ModelUtils] = None,
    ) -> None:
        """Initialize an EquipmentHandler instance."""
        self._hardware_api = hardware_api
        self._state_store = state_store
        self._labware_data_provider = labware_data_provider or LabwareDataProvider()
        self._model_utils = model_utils or ModelUtils()

    async def load_labware(
        self,
        load_name: str,
        namespace: str,
        version: int,
        location: LabwareLocation,
        labware_id: Optional[str],
    ) -> LoadedLabwareData:
        """Load labware by assigning an identifier and pulling required data.

        Args:
            load_name: The labware's load name.
            namespace: The namespace.
            version: Version
            location: The deck location at which labware is placed.
            labware_id: An optional identifier to assign the labware. If None, an
                identifier will be generated.

        Returns:
            A LoadedLabwareData object.
        """
        labware_id = labware_id if labware_id else self._model_utils.generate_id()

        try:
            # Try to use existing definition in state.
            definition = self._state_store.labware.get_definition_by_uri(
                uri_from_details(
                    load_name=load_name,
                    namespace=namespace,
                    version=version,
                )
            )
        except LabwareDefinitionDoesNotExistError:
            definition = await self._labware_data_provider.get_labware_definition(
                load_name=load_name,
                namespace=namespace,
                version=version,
            )

        calibration = await self._labware_data_provider.get_labware_calibration(
            definition=definition,
            location=location,
        )

        return LoadedLabwareData(
            labware_id=labware_id,
            definition=definition,
            calibration=calibration,
        )

    async def load_pipette(
        self,
        pipette_name: PipetteName,
        mount: MountType,
        pipette_id: Optional[str],
    ) -> LoadedPipetteData:
        """Ensure the requested pipette is attached.

        Args:
            pipette_name: The pipette name.
            mount: The mount on which pipette must be attached.
            pipette_id: An optional identifier to assign the pipette. If None, an
                identifier will be generated.

        Returns:
            A LoadedPipetteData object.
        """
        other_mount = mount.other_mount()
        other_pipette = self._state_store.pipettes.get_by_mount(other_mount)

        cache_request = {mount.to_hw_mount(): pipette_name}
        if other_pipette is not None:
            cache_request[other_mount.to_hw_mount()] = other_pipette.pipetteName

        # TODO(mc, 2020-10-18): calling `cache_instruments` mirrors the
        # behavior of protocol_context.load_instrument, and is used here as a
        # pipette existence check
        # TODO(mc, 2021-04-16): reconcile PipetteName enum with PipetteName union
        try:
            await self._hardware_api.cache_instruments(cache_request)  # type: ignore[arg-type]  # noqa: E501
        except RuntimeError as e:
            raise FailedToLoadPipetteError(str(e)) from e

        pipette_id = pipette_id or self._model_utils.generate_id()

        return LoadedPipetteData(pipette_id=pipette_id)



"""Command execution module."""

from .create_queue_worker import create_queue_worker
from .command_executor import CommandExecutor
from .queue_worker import QueueWorker
from .equipment import EquipmentHandler, LoadedLabwareData, LoadedPipetteData
from .movement import MovementHandler
from .pipetting import PipettingHandler
from .run_control import RunControlHandler


__all__ = [
    "CommandExecutor",
    "QueueWorker",
    "EquipmentHandler",
    "LoadedLabwareData",
    "LoadedPipetteData",
    "MovementHandler",
    "PipettingHandler",
    "RunControlHandler",
    "QueueWorker",
    "create_queue_worker",
]



"""The handler of a driver client connection."""

import asyncio
import logging

from opentrons.hardware_control.emulation.abstract_emulator import AbstractEmulator

logger = logging.getLogger(__name__)


class ConnectionHandler:
    """Responsible for reading data and routing it to an emulator."""

    def __init__(self, emulator: AbstractEmulator):
        """Construct with an emulator."""
        self._emulator = emulator

    async def __call__(
        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
    ) -> None:
        """New connection callback."""
        emulator_name = self._emulator.__class__.__name__
        logger.debug("%s Connected.", emulator_name)
        while True:
            line = await reader.readuntil(self._emulator.get_terminator())
            logger.debug("%s Received: %s", emulator_name, line)
            try:
                response = self._emulator.handle(line.decode().strip())
                if response:
                    response = f"{response}\r\n"
                    logger.debug("%s Sending: %s", emulator_name, response)
                    writer.write(response.encode())
            except Exception as e:
                logger.exception("%s exception", emulator_name)
                writer.write(f"Error: {str(e)}\r\n".encode())

            writer.write(self._emulator.get_ack())
            await writer.drain()



"""Basic pipette data state and store."""
from __future__ import annotations
from dataclasses import dataclass, replace
from typing import Dict, List, Mapping, Optional

from opentrons.hardware_control.dev_types import PipetteDict
from opentrons.types import MountType, Mount as HwMount

from .. import errors
from ..types import LoadedPipette

from ..commands import (
    Command,
    LoadPipetteResult,
    AspirateResult,
    DispenseResult,
    MoveToWellResult,
    PickUpTipResult,
    DropTipResult,
)
from .actions import Action, UpdateCommandAction
from .abstract_store import HasState, HandlesActions


@dataclass(frozen=True)
class HardwarePipette:
    """Hardware pipette data."""

    mount: HwMount
    config: PipetteDict


@dataclass(frozen=True)
class CurrentWell:
    """The latest well that the robot has accessed."""

    pipette_id: str
    labware_id: str
    well_name: str


@dataclass(frozen=True)
class PipetteState:
    """Basic labware data state and getter methods."""

    pipettes_by_id: Dict[str, LoadedPipette]
    aspirated_volume_by_id: Dict[str, float]
    current_well: Optional[CurrentWell]


class PipetteStore(HasState[PipetteState], HandlesActions):
    """Pipette state container."""

    _state: PipetteState

    def __init__(self) -> None:
        """Initialize a PipetteStore and its state."""
        self._state = PipetteState(
            pipettes_by_id={},
            aspirated_volume_by_id={},
            current_well=None,
        )

    def handle_action(self, action: Action) -> None:
        """Modify state in reaction to an action."""
        if isinstance(action, UpdateCommandAction):
            self._handle_command(action.command)

    def _handle_command(self, command: Command) -> None:
        if isinstance(
            command.result,
            (
                MoveToWellResult,
                PickUpTipResult,
                DropTipResult,
                AspirateResult,
                DispenseResult,
            ),
        ):
            self._state = replace(
                self._state,
                current_well=CurrentWell(
                    pipette_id=command.data.pipetteId,
                    labware_id=command.data.labwareId,
                    well_name=command.data.wellName,
                ),
            )

        if isinstance(command.result, LoadPipetteResult):
            pipette_id = command.result.pipetteId
            pipettes_by_id = self._state.pipettes_by_id.copy()
            aspirated_volume_by_id = self._state.aspirated_volume_by_id.copy()

            pipettes_by_id[pipette_id] = LoadedPipette(
                id=pipette_id,
                pipetteName=command.data.pipetteName,
                mount=command.data.mount,
            )
            aspirated_volume_by_id[pipette_id] = 0

            self._state = replace(
                self._state,
                pipettes_by_id=pipettes_by_id,
                aspirated_volume_by_id=aspirated_volume_by_id,
            )

        elif isinstance(command.result, AspirateResult):
            pipette_id = command.data.pipetteId
            aspirated_volume_by_id = self._state.aspirated_volume_by_id.copy()

            previous_volume = self._state.aspirated_volume_by_id[pipette_id]
            next_volume = previous_volume + command.result.volume
            aspirated_volume_by_id[pipette_id] = next_volume

            self._state = replace(
                self._state,
                aspirated_volume_by_id=aspirated_volume_by_id,
            )

        elif isinstance(command.result, DispenseResult):
            pipette_id = command.data.pipetteId
            aspirated_volume_by_id = self._state.aspirated_volume_by_id.copy()

            previous_volume = self._state.aspirated_volume_by_id[pipette_id]
            next_volume = max(0, previous_volume - command.result.volume)
            aspirated_volume_by_id[pipette_id] = next_volume

            self._state = replace(
                self._state,
                aspirated_volume_by_id=aspirated_volume_by_id,
            )


class PipetteView(HasState[PipetteState]):
    """Read-only view of computed pipettes state."""

    _state: PipetteState

    def __init__(self, state: PipetteState) -> None:
        """Initialize the view with its backing state value."""
        self._state = state

    def get(self, pipette_id: str) -> LoadedPipette:
        """Get pipette data by the pipette's unique identifier."""
        try:
            return self._state.pipettes_by_id[pipette_id]
        except KeyError:
            raise errors.PipetteDoesNotExistError(f"Pipette {pipette_id} not found.")

    def get_all(self) -> List[LoadedPipette]:
        """Get a list of all pipette entries in state."""
        return list(self._state.pipettes_by_id.values())

    def get_by_mount(self, mount: MountType) -> Optional[LoadedPipette]:
        """Get pipette data by the pipette's mount."""
        for pipette in self._state.pipettes_by_id.values():
            if pipette.mount == mount:
                return pipette
        return None

    def get_hardware_pipette(
        self,
        pipette_id: str,
        attached_pipettes: Mapping[HwMount, Optional[PipetteDict]],
    ) -> HardwarePipette:
        """Get a pipette's hardware configuration and state by ID."""
        pipette_data = self.get(pipette_id)
        pipette_name = pipette_data.pipetteName
        mount = pipette_data.mount

        hw_mount = mount.to_hw_mount()
        hw_config = attached_pipettes[hw_mount]

        if hw_config is None:
            raise errors.PipetteNotAttachedError(f"No pipetted attached on {mount}")
        # TODO(mc, 2020-11-12): support hw_pipette.act_as
        elif hw_config["name"] != pipette_name:
            raise errors.PipetteNotAttachedError(
                f"Found {hw_config['name']} on {mount}, "
                f"but {pipette_id} is a {pipette_name}"
            )

        return HardwarePipette(mount=hw_mount, config=hw_config)

    def get_current_well(self) -> Optional[CurrentWell]:
        """Get the last accessed well and which pipette accessed it."""
        return self._state.current_well

    def get_aspirated_volume(self, pipette_id: str) -> float:
        """Get the currently aspirated volume of a pipette by ID."""
        try:
            return self._state.aspirated_volume_by_id[pipette_id]
        except KeyError:
            raise errors.PipetteDoesNotExistError(
                f"Pipette {pipette_id} not found; unable to get current volume."
            )

    def get_is_ready_to_aspirate(
        self,
        pipette_id: str,
        pipette_config: PipetteDict,
    ) -> bool:
        """Get whether a pipette is ready to aspirate."""
        return (
            self.get_aspirated_volume(pipette_id) > 0
            or pipette_config["ready_to_aspirate"]
        )



"""
opentrons.system.log_control: functions for talking to syslog and journald

This is the implementation of the endpoints in
:py:mod:`opentrons.server.endpoints.logs` and friends.
"""
import asyncio
import logging
import subprocess
from typing import Tuple


LOG = logging.getLogger(__name__)

MAX_RECORDS = 100000
DEFAULT_RECORDS = 50000


async def get_records_dumb(selector: str, records: int, mode: str) -> bytes:
    """Dump the log files.

    :param selector: The syslog selector to limit responses to
    :param records: The maximum number of records to print
    :param mode: A journalctl dump mode. Should be either "short" or "json".
    """
    proc = await asyncio.create_subprocess_exec(
        "journalctl",
        "--no-pager",
        "-t",
        selector,
        "-n",
        str(records),
        "-o",
        mode,
        "-a",
        stdout=subprocess.PIPE,
    )
    stdout, _ = await proc.communicate()
    return stdout


async def set_syslog_level(level: str) -> Tuple[int, str, str]:
    """
    Set the minimum level for which logs will be sent upstream via syslog-ng.

    This is the function that actually does the work for
    :py:meth:`set_syslog_level_handler`.

    Similar to :py:meth:`opentrons.server.endpoints.settings.set_log_level`,
    the level should be a python log level like "debug", "info", "warning", or
    "error". If it is null, sets the minimum log level to emergency which we
    do not log at since there's not really a matching level in python logging,
    which effectively disables log upstreaming.

    :returns tuple(int, str, str): The error code, stdout, and stderr from
                                   ``syslog-ng-ctl``. ``0`` is success,
                                   anything else is failure
    """
    with open("/var/lib/syslog-ng/min-level", "w") as ml:
        ml.write(level)
    proc = await asyncio.create_subprocess_exec(
        "syslog-ng-ctl",
        "reload",
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )
    stdout, stderr = await proc.communicate()
    if proc.returncode is None:
        snc_reload_result = -1
    else:
        snc_reload_result: int = proc.returncode  # type: ignore
    return snc_reload_result, stdout.decode(), stderr.decode()



"""Depth first search.

Search a generic graph down to its leaf
nodes first before back-tracking up the tree.
"""

from typing import List, Set, Generic

from .graph import Graph
from .types import VertexLike, VertexName


class DFS(Generic[VertexName]):
    """Depth first search class.

    This class will build a graph object and then
    perform a depth first search on the graph.
    """

    def __init__(self, graph: List[VertexLike]) -> None:
        """DFS Initializer.

        :param graph: A list of nodes you wish to add to
        the graph.
        """
        self._graph = Graph.build(graph)

    @property
    def graph(self) -> Graph:
        """DFS property: graph.

        :returns: the graph object in which
        dfs is being performed on.
        """
        return self._graph

    def dfs(self) -> Set[VertexName]:
        """Depth first search.

        :returns: the set of visited vertices
        in depth first search order.
        """
        visited_vertices: Set[VertexName] = set()
        for node in self.graph.graph:
            if node not in visited_vertices:
                visited_vertices.add(node.name)
            for neighbor in node.neighbors:
                if neighbor not in visited_vertices:
                    visited_vertices.add(neighbor)
        return visited_vertices



import abc
import asyncio
import logging
import re
from pkg_resources import parse_version
from typing import Mapping, Optional
from opentrons.config import IS_ROBOT, ROBOT_FIRMWARE_DIR
from opentrons.hardware_control.util import use_or_initialize_loop
from opentrons.drivers.rpi_drivers.types import USBPort
from ..execution_manager import ExecutionManager
from .types import BundledFirmware, UploadFunction, InterruptCallback, LiveData

mod_log = logging.getLogger(__name__)


class AbstractModule(abc.ABC):
    """Defines the common methods of a module."""

    @classmethod
    @abc.abstractmethod
    async def build(
        cls,
        port: str,
        usb_port: USBPort,
        execution_manager: ExecutionManager,
        interrupt_callback: InterruptCallback = None,
        simulating: bool = False,
        loop: asyncio.AbstractEventLoop = None,
        sim_model: str = None,
        **kwargs,
    ) -> "AbstractModule":
        """Modules should always be created using this factory.

        This lets the (perhaps blocking) work of connecting to and initializing
        a module be in a place that can be async.
        """
        pass

    def __init__(
        self,
        port: str,
        usb_port: USBPort,
        execution_manager: ExecutionManager,
        simulating: bool = False,
        loop: asyncio.AbstractEventLoop = None,
        sim_model: str = None,
    ) -> None:
        self._port = port
        self._usb_port = usb_port
        self._loop = use_or_initialize_loop(loop)
        self._execution_manager = execution_manager
        self._bundled_fw: Optional[BundledFirmware] = self.get_bundled_fw()

    def get_bundled_fw(self) -> Optional[BundledFirmware]:
        """Get absolute path to bundled version of module fw if available."""
        if not IS_ROBOT:
            return None
        name_to_fw_file_prefix = {
            "tempdeck": "temperature-module",
            "magdeck": "magnetic-module",
        }
        name = self.name()
        file_prefix = name_to_fw_file_prefix.get(name, name)

        MODULE_FW_RE = re.compile(f"^{file_prefix}@v(.*)[.](hex|bin)$")
        for fw_resource in ROBOT_FIRMWARE_DIR.iterdir():  # type: ignore
            matches = MODULE_FW_RE.search(fw_resource.name)
            if matches:
                return BundledFirmware(version=matches.group(1), path=fw_resource)

        mod_log.info(f"no available fw file found for: {file_prefix}")
        return None

    def has_available_update(self) -> bool:
        """Return whether a newer firmware file is available"""
        if self.device_info and self._bundled_fw:
            device_version = parse_version(self.device_info["version"])
            available_version = parse_version(self._bundled_fw.version)
            return available_version > device_version
        return False

    async def wait_for_is_running(self):
        if not self.is_simulated:
            await self._execution_manager.wait_for_is_running()

    async def make_cancellable(self, task: asyncio.Task):
        self._execution_manager.register_cancellable_task(task)

    @abc.abstractmethod
    def deactivate(self):
        """Deactivate the module."""
        pass

    @property
    @abc.abstractmethod
    def status(self) -> str:
        """Return some string describing status."""
        pass

    @property
    @abc.abstractmethod
    def device_info(self) -> Mapping[str, str]:
        """Return a dict of the module's static information (serial, etc)"""
        pass

    @property
    @abc.abstractmethod
    def live_data(self) -> LiveData:
        """Return a dict of the module's dynamic information"""
        pass

    @property
    @abc.abstractmethod
    def is_simulated(self) -> bool:
        """True if >this is a simulated module."""
        pass

    @property
    def port(self) -> str:
        """The virtual port where the module is connected."""
        return self._port

    @property
    def usb_port(self) -> USBPort:
        """The physical port where the module is connected."""
        return self._usb_port

    @abc.abstractmethod
    async def prep_for_update(self) -> str:
        """Prepare for an update.

        By the time this coroutine completes, the hardware should be ready
        to take an update. This implicitly tears down the module instance;
        it does not need to be either working or recoverable after this
        coroutine completes.

        :returns str: The port we're running on.
        """
        pass

    @property
    def bundled_fw(self):
        return self._bundled_fw

    @abc.abstractmethod
    def model(self) -> str:
        """A name for this specific module, matching module defs"""
        pass

    @classmethod
    @abc.abstractmethod
    def name(cls) -> str:
        """A shortname used for looking up firmware, among other things"""
        pass

    @classmethod
    @abc.abstractmethod
    def bootloader(cls) -> UploadFunction:
        """Method used to upload file to this module's bootloader."""
        pass

    async def cleanup(self) -> None:
        """Clean up the module instance.

        Clean up, i.e. stop pollers, disconnect serial, etc in preparation for
        object destruction.
        """
        pass



""" opentrons.calibration_storage.get: functions for grabing calibration

This module has functions that you can import to load robot or
labware calibration from its designated file location.
"""
import itertools
import logging
import json
import typing
from typing_extensions import Literal

from opentrons import config
from opentrons.types import Point, Mount

from . import types as local_types, file_operators as io, helpers, migration, modify

if typing.TYPE_CHECKING:
    from opentrons_shared_data.labware.dev_types import LabwareDefinition
    from opentrons_shared_data.pipette.dev_types import LabwareUri
    from .dev_types import CalibrationIndexDict, CalibrationDict


log = logging.getLogger(__name__)


def _format_calibration_type(
    data: "CalibrationDict",
) -> local_types.LabwareCalibrationTypes:
    offset = local_types.OffsetData(
        value=data["default"]["offset"], last_modified=data["default"]["lastModified"]
    )
    # TODO(6/16): Tip calibration no longer exists in
    # the labware calibraiton file. We should
    # have a follow-up PR to grab tip lengths
    # based on the loaded pips + labware
    return local_types.LabwareCalibrationTypes(
        offset=offset, tip_length=local_types.TipLengthData()
    )


def _format_parent(data: "CalibrationIndexDict") -> local_types.ParentOptions:
    # Since the slot is not saved and the data in the index is actually
    # the labware hash to aid lookup, we erase it here.
    options = local_types.ParentOptions(slot="")
    if data["module"]:
        options.module = data["module"]["parent"]
    return options


def get_all_calibrations() -> typing.List[local_types.CalibrationInformation]:
    """
    A helper function that will list all of the given calibrations
    in a succinct way.

    :return: A list of dictionary objects representing all of the
    labware calibration files found on the robot.
    """
    all_calibrations: typing.List[local_types.CalibrationInformation] = []
    offset_path = config.get_opentrons_path("labware_calibration_offsets_dir_v2")
    index_path = offset_path / "index.json"
    if not index_path.exists():
        return all_calibrations

    migration.check_index_version(index_path)
    index_file = io.read_cal_file(str(index_path))
    calibration_index = index_file.get("data", {})
    for key, data in calibration_index.items():
        cal_path = offset_path / f"{key}.json"
        if cal_path.exists():
            try:
                cal_blob = io.read_cal_file(str(cal_path))
            except json.JSONDecodeError:
                log.error(
                    f"Skipping corrupt calibration file (bad JSON): {str(cal_path)}"
                )
                continue
            calibration = _format_calibration_type(cal_blob)  # type: ignore
            try:
                all_calibrations.append(
                    local_types.CalibrationInformation(
                        calibration=calibration,
                        parent=_format_parent(data),
                        labware_id=key,
                        uri=data["uri"],
                    )
                )
            except (KeyError, ValueError):
                log.exception(
                    f"Skipping corrupt calibration file (bad data) {str(cal_path)}"
                )
                continue
    return all_calibrations


def _get_tip_length_data(
    pip_id: str, labware_hash: str, labware_load_name: str, labware_uri: "LabwareUri"
) -> local_types.TipLengthCalibration:
    try:
        pip_tip_length_path = config.get_tip_length_cal_path() / f"{pip_id}.json"
        tip_rack_data = io.read_cal_file(str(pip_tip_length_path))
        tip_length_info = tip_rack_data[labware_hash]
        return local_types.TipLengthCalibration(
            tip_length=tip_length_info["tipLength"],
            pipette=pip_id,
            tiprack=labware_hash,
            last_modified=tip_length_info["lastModified"],
            source=_get_calibration_source(tip_length_info),
            status=_get_calibration_status(tip_length_info),
            uri=labware_uri,
        )
    except (FileNotFoundError, KeyError, json.JSONDecodeError):
        raise local_types.TipLengthCalNotFound(
            f"Tip length of {labware_load_name} has not been "
            f"calibrated for this pipette: {pip_id} and cannot"
            "be loaded"
        )


def get_labware_calibration(
    lookup_path: local_types.StrPath,
    definition: "LabwareDefinition",
    parent: str = "",
    slot: str = "",
) -> Point:
    """
    Find the delta of a given labware, if it exists.

    :param lookup_path: short path to the labware calibration
    :return: A point which represents the delta from well A1 origin of
    a labware
    """
    offset_path = config.get_opentrons_path("labware_calibration_offsets_dir_v2")
    offset = Point(0, 0, 0)
    labware_path = offset_path / lookup_path
    if labware_path.exists():
        modify.add_existing_labware_to_index_file(definition, parent, slot)
        migration.check_index_version(offset_path / "index.json")
        calibration_data = io.read_cal_file(str(labware_path))
        offset_array = calibration_data["default"]["offset"]
        offset = Point(x=offset_array[0], y=offset_array[1], z=offset_array[2])
    return offset


def load_tip_length_calibration(
    pip_id: str, definition: "LabwareDefinition"
) -> local_types.TipLengthCalibration:
    """
    Function used to grab the current tip length associated
    with a particular tiprack.

    :param pip_id: pipette you are using
    :param definition: full definition of the tiprack
    """
    labware_hash = helpers.hash_labware_def(definition)
    labware_uri = helpers.uri_from_definition(definition)
    load_name = definition["parameters"]["loadName"]
    return _get_tip_length_data(
        pip_id=pip_id,
        labware_hash=labware_hash,
        labware_load_name=load_name,
        labware_uri=labware_uri,
    )


def get_all_tip_length_calibrations() -> typing.List[local_types.TipLengthCalibration]:
    """
    A helper function that will list all of the tip length calibrations.

    :return: A list of dictionary objects representing all of the
    tip length calibration files found on the robot.
    """
    all_calibrations: typing.List[local_types.TipLengthCalibration] = []
    tip_length_dir = config.get_opentrons_path("tip_length_calibration_dir")
    index_path = tip_length_dir / "index.json"
    if not index_path.exists():
        return all_calibrations

    index_file = io.read_cal_file(str(index_path))
    unique_pips = set(itertools.chain(*index_file.values()))
    for pip in unique_pips:
        cal_path = tip_length_dir / f"{pip}.json"
        if cal_path.exists():
            try:
                data = io.read_cal_file(str(cal_path))
            except json.JSONDecodeError:
                log.error(
                    f"Skipping corrupt calibration file (bad json): {str(cal_path)}"
                )
                continue
            for tiprack, info in data.items():
                all_calibrations.append(
                    local_types.TipLengthCalibration(
                        tip_length=info["tipLength"],
                        pipette=pip,
                        tiprack=tiprack,
                        last_modified=info["lastModified"],
                        source=_get_calibration_source(info),
                        status=_get_calibration_status(info),
                        uri=_get_tip_rack_uri(info),
                    )
                )
    return all_calibrations


def _get_calibration_source(data: typing.Dict) -> local_types.SourceType:
    if "source" not in data.keys():
        return local_types.SourceType.unknown
    else:
        return local_types.SourceType[data["source"]]


def _get_calibration_status(data: typing.Dict) -> local_types.CalibrationStatus:
    if "status" not in data.keys():
        return local_types.CalibrationStatus()
    else:
        return local_types.CalibrationStatus(**data["status"])


def _get_tip_rack_uri(data: typing.Dict) -> typing.Union["LabwareUri", Literal[""]]:
    if "uri" not in data.keys():
        # We cannot reverse look-up a labware definition using a hash
        # so we must return an empty string if no uri is found.
        return ""
    else:
        return typing.cast("LabwareUri", data["uri"])


def get_robot_deck_attitude() -> typing.Optional[local_types.DeckCalibration]:
    robot_dir = config.get_opentrons_path("robot_calibration_dir")
    gantry_path = robot_dir / "deck_calibration.json"
    if gantry_path.exists():
        try:
            data = io.read_cal_file(gantry_path)
        except json.JSONDecodeError:
            log.error(
                f"Skipping corrupt calibration file (bad json): {str(gantry_path)}"
            )
            return None
        try:
            return local_types.DeckCalibration(
                attitude=data["attitude"],
                source=_get_calibration_source(data),
                pipette_calibrated_with=data["pipette_calibrated_with"],
                tiprack=data["tiprack"],
                last_modified=data["last_modified"],
                status=_get_calibration_status(data),
            )
        except Exception:
            return None
    else:
        return None


def get_pipette_offset(
    pip_id: str, mount: Mount
) -> typing.Optional[local_types.PipetteOffsetByPipetteMount]:
    pip_dir = config.get_opentrons_path("pipette_calibration_dir")
    offset_path = pip_dir / mount.name.lower() / f"{pip_id}.json"
    if offset_path.exists():
        try:
            data = io.read_cal_file(offset_path)
        except json.JSONDecodeError:
            log.error(
                f"Skipping corrupt calibration file (bad json): {str(offset_path)}"
            )
            return None
        assert "offset" in data.keys(), "Not valid pipette calibration data"
        return local_types.PipetteOffsetByPipetteMount(
            offset=data["offset"],
            source=_get_calibration_source(data),
            tiprack=data["tiprack"],
            uri=data["uri"],
            last_modified=data["last_modified"],
            status=_get_calibration_status(data),
        )
    else:
        return None


def get_all_pipette_offset_calibrations() -> typing.List[
    local_types.PipetteOffsetCalibration
]:
    """
    A helper function that will list all of the pipette offset
    calibrations.

    :return: A list of dictionary objects representing all of the
    pipette offset calibration files found on the robot.
    """
    all_calibrations: typing.List[local_types.PipetteOffsetCalibration] = []
    pip_dir = config.get_opentrons_path("pipette_calibration_dir")
    index_path = pip_dir / "index.json"
    if not index_path.exists():
        return all_calibrations

    index_file = io.read_cal_file(str(index_path))
    for mount_key, pips in index_file.items():
        for pip in pips:
            cal_path = pip_dir / mount_key / f"{pip}.json"
            if cal_path.exists():
                try:
                    data = io.read_cal_file(str(cal_path))
                except json.JSONDecodeError:
                    log.error(
                        "Skipping corrupt calibration file (bad json): "
                        + f"{str(cal_path)}"
                    )
                    continue
                try:
                    all_calibrations.append(
                        local_types.PipetteOffsetCalibration(
                            pipette=pip,
                            mount=mount_key,
                            offset=data["offset"],
                            tiprack=data["tiprack"],
                            uri=data["uri"],
                            last_modified=data["last_modified"],
                            source=_get_calibration_source(data),
                            status=_get_calibration_status(data),
                        )
                    )
                except (KeyError, ValueError):
                    log.exception(
                        "Skipping corrupt calibration file (bad data): "
                        + f"{str(cal_path)}"
                    )
                    continue
    return all_calibrations


def get_custom_tiprack_definition_for_tlc(labware_uri: str) -> "LabwareDefinition":
    """
    Return the custom tiprack definition saved in the custom tiprack directory
    during tip length calibration
    """
    custom_tiprack_dir = config.get_custom_tiprack_def_path()
    custom_tiprack_path = custom_tiprack_dir / f"{labware_uri}.json"
    try:
        with open(custom_tiprack_path, "rb") as f:
            return json.loads(f.read().decode("utf-8"))
    except FileNotFoundError:
        raise FileNotFoundError(
            f"Custom tiprack {labware_uri} not found in the custom tiprack"
            "directory on the robot. Please recalibrate tip length and "
            "pipette offset with this tiprack before performing calibration "
            "health check."
        )



"""Basic labware data state and store."""
from __future__ import annotations

import re
from collections import defaultdict
from dataclasses import dataclass, replace
from typing import Dict, List, Sequence

from opentrons_shared_data.deck.dev_types import DeckDefinitionV2, SlotDefV2
from opentrons_shared_data.labware.constants import WELL_NAME_PATTERN
from opentrons_shared_data.pipette.dev_types import LabwareUri

from opentrons.types import DeckSlotName, Point
from opentrons.protocols.models import LabwareDefinition, WellDefinition
from opentrons.calibration_storage.helpers import uri_from_details

from .. import errors
from ..resources import DeckFixedLabware
from ..commands import Command, LoadLabwareResult, AddLabwareDefinitionResult
from ..types import CalibrationOffset, LabwareLocation, LoadedLabware, Dimensions
from .actions import Action, UpdateCommandAction
from .abstract_store import HasState, HandlesActions


@dataclass(frozen=True)
class LabwareState:
    """State of all loaded labware resources."""

    labware_by_id: Dict[str, LoadedLabware]
    calibrations_by_id: Dict[str, CalibrationOffset]
    definitions_by_uri: Dict[str, LabwareDefinition]
    deck_definition: DeckDefinitionV2


class LabwareStore(HasState[LabwareState], HandlesActions):
    """Labware state container."""

    _state: LabwareState

    def __init__(
        self,
        deck_definition: DeckDefinitionV2,
        deck_fixed_labware: Sequence[DeckFixedLabware],
    ) -> None:
        """Initialize a labware store and its state."""
        definitions_by_uri: Dict[str, LabwareDefinition] = {
            uri_from_details(
                load_name=fixed_labware.definition.parameters.loadName,
                namespace=fixed_labware.definition.namespace,
                version=fixed_labware.definition.version,
            ): fixed_labware.definition
            for fixed_labware in deck_fixed_labware
        }
        labware_by_id = {
            fixed_labware.labware_id: LoadedLabware(
                id=fixed_labware.labware_id,
                location=fixed_labware.location,
                loadName=fixed_labware.definition.parameters.loadName,
                definitionUri=uri_from_details(
                    load_name=fixed_labware.definition.parameters.loadName,
                    namespace=fixed_labware.definition.namespace,
                    version=fixed_labware.definition.version,
                ),
            )
            for fixed_labware in deck_fixed_labware
        }
        calibrations_by_id = {
            fixed_labware.labware_id: CalibrationOffset(x=0, y=0, z=0)
            for fixed_labware in deck_fixed_labware
        }

        self._state = LabwareState(
            definitions_by_uri=definitions_by_uri,
            calibrations_by_id=calibrations_by_id,
            labware_by_id=labware_by_id,
            deck_definition=deck_definition,
        )

    def handle_action(self, action: Action) -> None:
        """Modify state in reaction to an action."""
        if isinstance(action, UpdateCommandAction):
            self._handle_command(action.command)

    def _handle_command(self, command: Command) -> None:
        """Modify state in reaction to a command."""
        if isinstance(command.result, LoadLabwareResult):
            labware_id = command.result.labwareId
            definition_uri = uri_from_details(
                namespace=command.result.definition.namespace,
                load_name=command.result.definition.parameters.loadName,
                version=command.result.definition.version,
            )
            labware_by_id = self._state.labware_by_id.copy()
            calibrations_by_id = self._state.calibrations_by_id.copy()
            definitions_by_uri = self._state.definitions_by_uri.copy()

            definitions_by_uri[definition_uri] = command.result.definition
            calibrations_by_id[labware_id] = command.result.calibration
            labware_by_id[labware_id] = LoadedLabware(
                id=labware_id,
                location=command.data.location,
                loadName=command.result.definition.parameters.loadName,
                definitionUri=definition_uri,
            )

            self._state = replace(
                self._state,
                labware_by_id=labware_by_id,
                calibrations_by_id=calibrations_by_id,
                definitions_by_uri=definitions_by_uri,
            )

        elif isinstance(command.result, AddLabwareDefinitionResult):
            definition_uri = uri_from_details(
                namespace=command.result.namespace,
                load_name=command.result.loadName,
                version=command.result.version,
            )
            definitions_by_uri = self._state.definitions_by_uri.copy()
            definitions_by_uri[definition_uri] = command.data.definition
            self._state = replace(self._state, definitions_by_uri=definitions_by_uri)


class LabwareView(HasState[LabwareState]):
    """Read-only labware state view."""

    _state: LabwareState

    def __init__(self, state: LabwareState) -> None:
        """Initialize the computed view of labware state.

        Arguments:
            state: Labware state dataclass used for all calculations.
        """
        self._state = state

    def get(self, labware_id: str) -> LoadedLabware:
        """Get labware data by the labware's unique identifier."""
        try:
            return self._state.labware_by_id[labware_id]
        except KeyError:
            raise errors.LabwareDoesNotExistError(f"Labware {labware_id} not found.")

    def get_definition(self, labware_id: str) -> LabwareDefinition:
        """Get labware definition by the labware's unique identifier."""
        return self.get_definition_by_uri(
            LabwareUri(self.get(labware_id).definitionUri)
        )

    def get_deck_definition(self) -> DeckDefinitionV2:
        """Get the current deck definition."""
        return self._state.deck_definition

    def get_slot_definition(self, slot: DeckSlotName) -> SlotDefV2:
        """Get the definition of a slot in the deck."""
        deck_def = self.get_deck_definition()

        for slot_def in deck_def["locations"]["orderedSlots"]:
            if slot_def["id"] == str(slot):
                return slot_def

        raise errors.SlotDoesNotExistError(
            f"Slot ID {slot} does not exist in deck {deck_def['otId']}"
        )

    def get_slot_position(self, slot: DeckSlotName) -> Point:
        """Get the position of a deck slot."""
        slot_def = self.get_slot_definition(slot)
        position = slot_def["position"]

        return Point(x=position[0], y=position[1], z=position[2])

    def get_definition_by_uri(self, uri: LabwareUri) -> LabwareDefinition:
        """Get the labware definition matching loadName namespace and version."""
        try:
            return self._state.definitions_by_uri[uri]
        except KeyError:
            raise errors.LabwareDefinitionDoesNotExistError(
                f"Labware definition for matching {uri} not found."
            )

    def get_location(self, labware_id: str) -> LabwareLocation:
        """Get labware location by the labware's unique identifier."""
        return self.get(labware_id).location

    def get_all(self) -> List[LoadedLabware]:
        """Get a list of all labware entries in state."""
        return list(self._state.labware_by_id.values())

    def get_has_quirk(self, labware_id: str, quirk: str) -> bool:
        """Get if a labware has a certain quirk."""
        return quirk in self.get_quirks(labware_id=labware_id)

    def get_quirks(self, labware_id: str) -> List[str]:
        """Get a labware's quirks."""
        definition = self.get_definition(labware_id)
        return definition.parameters.quirks or []

    def get_well_definition(
        self,
        labware_id: str,
        well_name: str,
    ) -> WellDefinition:
        """Get a well's definition by labware and well identifier."""
        definition = self.get_definition(labware_id)

        try:
            return definition.wells[well_name]
        except KeyError:
            raise errors.WellDoesNotExistError(
                f"{well_name} does not exist in {labware_id}."
            )

    def get_wells(self, labware_id: str) -> List[str]:
        """Get labware wells as a list of well names."""
        definition = self.get_definition(labware_id=labware_id)
        wells = list()
        for col in definition.ordering:
            for well_name in col:
                wells.append(well_name)
        return wells

    def get_well_columns(self, labware_id: str) -> Dict[str, List[str]]:
        """Get well columns."""
        definition = self.get_definition(labware_id=labware_id)
        wells_by_cols = defaultdict(list)
        for i, col in enumerate(definition.ordering):
            wells_by_cols[f"{i+1}"] = col
        return wells_by_cols

    def get_well_rows(self, labware_id: str) -> Dict[str, List[str]]:
        """Get well rows."""
        definition = self.get_definition(labware_id=labware_id)
        wells_by_rows = defaultdict(list)
        pattern = re.compile(WELL_NAME_PATTERN, re.X)
        for col in definition.ordering:
            for well_name in col:
                match = pattern.match(well_name)
                assert match, f"Well name did not match pattern {pattern}"
                wells_by_rows[match.group(1)].append(well_name)
        return wells_by_rows

    def get_tip_length(self, labware_id: str) -> float:
        """Get the tip length of a tip rack."""
        definition = self.get_definition(labware_id)
        if definition.parameters.tipLength is None:
            raise errors.LabwareIsNotTipRackError(
                f"Labware {labware_id} has no tip length defined."
            )
        return definition.parameters.tipLength

    def get_definition_uri(self, labware_id: str) -> str:
        """Get a labware's definition URI."""
        return self.get(labware_id).definitionUri

    def is_tiprack(self, labware_id: str) -> bool:
        """Get whether labware is a tiprack."""
        definition = self.get_definition(labware_id)
        return definition.parameters.isTiprack

    def get_load_name(self, labware_id: str) -> str:
        """Get the labware's load name."""
        definition = self.get_definition(labware_id)
        return definition.parameters.loadName

    def get_dimensions(self, labware_id: str) -> Dimensions:
        """Get the labware's dimensions."""
        definition = self.get_definition(labware_id)
        dims = definition.dimensions

        return Dimensions(
            x=dims.xDimension,
            y=dims.yDimension,
            z=dims.zDimension,
        )

    def get_calibration_offset(self, labware_id: str) -> CalibrationOffset:
        """Get the labware's calibration offset."""
        try:
            return self._state.calibrations_by_id[labware_id]
        except KeyError:
            raise errors.LabwareDoesNotExistError(f"Labware {labware_id} not found.")



from __future__ import annotations
import asyncio
import functools
import logging
from copy import copy
from typing import Any, cast, Callable, Optional, Union, Set, TypeVar
from typing_extensions import TypedDict, Literal, Final

from opentrons.config import feature_flags as ff
from opentrons.broker import Broker
from opentrons.types import Point, Mount, Location
from opentrons.protocol_api import labware
from opentrons.hardware_control import CriticalPoint, ThreadedAsyncLock
from opentrons.hardware_control.types import OutOfBoundsMove, MotionChecks
from opentrons.hardware_control.adapters import SynchronousAdapter

from .models import Container, Instrument
from .util import robot_is_busy, RobotBusy


log = logging.getLogger(__name__)

State = Union[Literal["moving"], Literal["ready"]]
VALID_STATES: Set[State] = {"moving", "ready"}


def _well0(cont: labware.Labware) -> labware.Well:
    return cont.wells()[0]


Func = TypeVar("Func", bound=Callable[..., Any])


def _home_if_first_call(func: Func) -> Func:
    """Decorator to make a function home if it is the first one called in
    this session."""

    @functools.wraps(func)
    def decorated(*args: Any, **kwargs: Any) -> Any:
        self = args[0]
        if not self._has_homed:
            log.info("this is the first calibration action, homing")
            self._hardware.home()
            self._has_homed = True
        return func(*args, **kwargs)

    return cast(Func, decorated)


class Message(TypedDict):
    topic: Literal["calibration"]
    name: Literal["state"]
    payload: CalibrationManager


class CalibrationManager(RobotBusy):
    """
    Serves endpoints that are primarily used in
    opentrons/app/ui/robot/api-client/client.js
    """

    TOPIC: Final = "calibration"

    def __init__(
        self,
        hardware: SynchronousAdapter,
        loop: asyncio.AbstractEventLoop = None,
        broker: Broker = None,
        lock: ThreadedAsyncLock = None,
    ) -> None:
        self._broker = broker or Broker()
        self._hardware = hardware
        self._loop = loop
        self.state = "ready"
        self._lock = lock or ThreadedAsyncLock()
        self._has_homed = False

    @property
    def busy_lock(self) -> ThreadedAsyncLock:
        return self._lock

    def _set_state(self, state: State) -> None:
        if state not in VALID_STATES:
            raise ValueError("State {0} not in {1}".format(state, VALID_STATES))
        self.state = state
        self._on_state_changed()

    @robot_is_busy
    @_home_if_first_call
    def tip_probe(self, instrument: Instrument) -> None:
        log.warning("Deprecated call to tip_probe, update app")
        self._set_state("ready")

    @robot_is_busy
    @_home_if_first_call
    def pick_up_tip(self, instrument: Instrument, container: Container) -> None:
        if not isinstance(container, Container):
            raise ValueError(
                "Invalid object type {0}. Expected models.Container".format(
                    type(container)
                )
            )

        inst = instrument._instrument
        log.info(
            "Picking up tip from {} in {} with {}".format(
                container.name, container.slot, instrument.name
            )
        )
        self._set_state("moving")
        with instrument._context.temp_connect(self._hardware):
            loc = _well0(container._container)
            instrument._context.location_cache = Location(
                self._hardware.gantry_position(
                    Mount[inst.mount.upper()],
                    critical_point=CriticalPoint.NOZZLE,
                    refresh=True,
                ),
                loc,
            )
            loc_leg = _well0(container._container)
            inst.pick_up_tip(loc_leg)
        self._set_state("ready")

    @robot_is_busy
    @_home_if_first_call
    def drop_tip(self, instrument: Instrument, container: Container) -> None:
        if not isinstance(container, Container):
            raise ValueError(
                "Invalid object type {0}. Expected models.Container".format(
                    type(container)
                )
            )

        inst = instrument._instrument
        log.info(
            "Dropping tip from {} in {} with {}".format(
                container.name, container.slot, instrument.name
            )
        )
        self._set_state("moving")
        with instrument._context.temp_connect(self._hardware):
            instrument._context.location_cache = None
            inst.drop_tip(_well0(container._container))
        self._set_state("ready")

    @robot_is_busy
    @_home_if_first_call
    def return_tip(self, instrument: Instrument) -> None:
        inst = instrument._instrument
        log.info("Returning tip from {}".format(instrument.name))
        self._set_state("moving")
        with instrument._context.temp_connect(self._hardware):
            instrument._context.location_cache = None
            inst.return_tip()
        self._set_state("ready")

    @robot_is_busy
    @_home_if_first_call
    def move_to_front(self, instrument: Instrument) -> None:
        """Public face of move_to_front"""
        self._move_to_front(instrument)

    def _move_to_front(self, instrument: Instrument) -> None:
        """Private move_to_front that can be called internally"""
        inst = instrument._instrument
        log.info("Moving {}".format(instrument.name))
        self._set_state("moving")
        current = self._hardware.gantry_position(
            Mount[inst.mount.upper()], critical_point=CriticalPoint.NOZZLE, refresh=True
        )
        dest = instrument._context.deck.position_for(5).point._replace(z=150)
        self._hardware.move_to(
            Mount[inst.mount.upper()], current, critical_point=CriticalPoint.NOZZLE
        )
        self._hardware.move_to(
            Mount[inst.mount.upper()],
            dest._replace(z=current.z),
            critical_point=CriticalPoint.NOZZLE,
        )
        self._hardware.move_to(
            Mount[inst.mount.upper()], dest, critical_point=CriticalPoint.NOZZLE
        )
        self._set_state("ready")

    @robot_is_busy
    @_home_if_first_call
    def move_to(self, instrument: Instrument, container: Container) -> None:
        if not isinstance(container, Container):
            raise ValueError(
                "Invalid object type {0}. Expected models.Container".format(
                    type(container)
                )
            )

        inst = instrument._instrument
        cont = container._container
        target = _well0(cont).top()

        log.info(
            "Moving {} to {} in {}".format(
                instrument.name, container.name, container.slot
            )
        )
        self._set_state("moving")

        with instrument._context.temp_connect(self._hardware):
            instrument._context.location_cache = None
            inst.move_to(target)

        self._set_state("ready")

    @robot_is_busy
    @_home_if_first_call
    def jog(self, instrument: Instrument, distance: float, axis: str) -> None:
        inst = instrument._instrument
        log.info("Jogging {} by {} in {}".format(instrument.name, distance, axis))
        self._set_state("moving")
        try:
            self._hardware.move_rel(
                Mount[inst.mount.upper()],
                Point(**{axis: distance}),
                check_bounds=MotionChecks.HIGH,
            )
        except OutOfBoundsMove:
            log.exception("Out of bounds jog")
        self._set_state("ready")

    @robot_is_busy
    @_home_if_first_call
    def home(self, instrument: Instrument) -> None:
        inst = instrument._instrument
        log.info("Homing {}".format(instrument.name))
        self._set_state("moving")
        with instrument._context.temp_connect(self._hardware):
            instrument._context.location_cache = None
            inst.home()
        self._set_state("ready")

    @robot_is_busy
    def home_all(self, instrument: Instrument) -> None:
        # NOTE: this only takes instrument as a param, because we need
        # its reference to the ProtocolContext. This is code smell that
        # will be removed once sessions are managed better
        log.info("Homing via Calibration Manager")
        self._set_state("moving")
        with instrument._context.temp_connect(self._hardware):
            instrument._context.home()
        self._set_state("ready")

    @robot_is_busy
    def update_container_offset(
        self, container: Container, instrument: Instrument
    ) -> None:
        inst = instrument._instrument
        log.info("Updating {} in {}".format(container.name, container.slot))
        if "centerMultichannelOnWells" in container._container.quirks:
            cp: Optional[CriticalPoint] = CriticalPoint.XY_CENTER
        else:
            cp = None
        here = self._hardware.gantry_position(
            Mount[inst.mount.upper()], critical_point=cp, refresh=True
        )
        # Reset calibration so we don’t actually calibrate the offset
        # relative to the old calibration
        container._container.set_calibration(Point(0, 0, 0))
        if ff.calibrate_to_bottom() and not (container._container.is_tiprack):
            orig = _well0(container._container).geometry.bottom()
        else:
            orig = _well0(container._container).geometry.top()
        delta = here - orig
        labware.save_calibration(container._container, delta)

    def _snapshot(self) -> Message:
        return {
            "topic": CalibrationManager.TOPIC,
            "name": "state",
            "payload": copy(self),
        }

    def _on_state_changed(self) -> None:
        self._broker.publish(CalibrationManager.TOPIC, self._snapshot())



"""Python file reading and parsing."""
import importlib.util
from types import ModuleType

from opentrons.protocol_api_experimental import ProtocolContext
from .protocol_file import ProtocolFile


class PythonProtocol:
    """A thin wrapper around the imported Python module."""

    def __init__(self, protocol_module: ModuleType) -> None:
        """Wrap the passed in protocol module."""
        self._protocol_module = protocol_module

    def run(self, context: ProtocolContext) -> None:
        """Call the protocol module's run method."""
        self._protocol_module.run(context)  # type: ignore[attr-defined]


class PythonFileReader:
    """A reader for Python protocol files.

    Gets a Python protocol's metadata (TODO) and run method.
    """

    @staticmethod
    def read(protocol_file: ProtocolFile) -> PythonProtocol:
        """Read a Python protocol as a `import`ed Python module."""
        # TODO(mc, 2021-06-30): better module name logic
        # TODO(mc, 2021-08-25): validate files list length before access
        spec = importlib.util.spec_from_file_location(
            name="protocol",
            location=protocol_file.files[0],
        )

        # TODO(mc, 2021-09-12): figure out why this would happen and raise
        # a well defined error accordingly
        assert spec is not None, "Unable to load module spec from file"

        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)  # type: ignore[union-attr]

        # TODO(mc, 2021-06-30): actually check that this module shape is good
        return PythonProtocol(protocol_module=module)



"""Motion state store and getters."""
from dataclasses import dataclass
from typing import List, Optional

from opentrons.types import MountType, Point
from opentrons.hardware_control.types import CriticalPoint
from opentrons.motion_planning import (
    MoveType,
    Waypoint,
    MotionPlanningError,
    get_waypoints,
)

from .. import errors
from ..types import WellLocation
from .labware import LabwareView
from .pipettes import PipetteView, CurrentWell
from .geometry import GeometryView


@dataclass(frozen=True)
class PipetteLocationData:
    """Pipette data used to determine the current gantry position."""

    mount: MountType
    critical_point: Optional[CriticalPoint]


class MotionView:
    """Complete motion planning state and getter methods."""

    _labware: LabwareView
    _pipette: PipetteView
    _geometry: GeometryView

    def __init__(
        self,
        labware_view: LabwareView,
        pipette_view: PipetteView,
        geometry_view: GeometryView,
    ) -> None:
        """Initialize a MotionState instance."""
        self._labware = labware_view
        self._pipettes = pipette_view
        self._geometry = geometry_view

    def get_pipette_location(
        self,
        pipette_id: str,
        current_well: Optional[CurrentWell] = None,
    ) -> PipetteLocationData:
        """Get the critical point of a pipette given the current location."""
        current_well = current_well or self._pipettes.get_current_well()
        pipette_data = self._pipettes.get(pipette_id)

        mount = pipette_data.mount
        critical_point = None

        # if the pipette was last used to move to a labware that requires
        # centering, set the critical point to XY_CENTER
        if (
            current_well is not None
            and current_well.pipette_id == pipette_id
            and self._labware.get_has_quirk(
                current_well.labware_id,
                "centerMultichannelOnWells",
            )
        ):
            critical_point = CriticalPoint.XY_CENTER

        return PipetteLocationData(mount=mount, critical_point=critical_point)

    def get_movement_waypoints(
        self,
        pipette_id: str,
        labware_id: str,
        well_name: str,
        well_location: Optional[WellLocation],
        origin: Point,
        origin_cp: Optional[CriticalPoint],
        max_travel_z: float,
        current_well: Optional[CurrentWell] = None,
    ) -> List[Waypoint]:
        """Get the movement waypoints from an origin to a given location."""
        location = current_well or self._pipettes.get_current_well()
        center_dest = self._labware.get_has_quirk(
            labware_id,
            "centerMultichannelOnWells",
        )

        dest = self._geometry.get_well_position(
            labware_id,
            well_name,
            well_location,
        )
        dest_cp = CriticalPoint.XY_CENTER if center_dest else None

        if (
            location is not None
            and pipette_id == location.pipette_id
            and labware_id == location.labware_id
        ):
            move_type = (
                MoveType.IN_LABWARE_ARC
                if well_name != location.well_name
                else MoveType.DIRECT
            )
            min_travel_z = self._geometry.get_labware_highest_z(labware_id)
        else:
            move_type = MoveType.GENERAL_ARC
            min_travel_z = self._geometry.get_all_labware_highest_z()

        try:
            # TODO(mc, 2021-01-08): inject `get_waypoints` via constructor
            return get_waypoints(
                move_type=move_type,
                origin=origin,
                origin_cp=origin_cp,
                dest=dest,
                dest_cp=dest_cp,
                min_travel_z=min_travel_z,
                max_travel_z=max_travel_z,
                xy_waypoints=[],
            )
        except MotionPlanningError as error:
            raise errors.FailedToPlanMoveError(str(error))



"""Protocol engine commands sub-state."""
from __future__ import annotations
from collections import OrderedDict
from dataclasses import dataclass, replace
from typing import List, Optional, Union

from ..commands import Command, CommandStatus
from ..errors import CommandDoesNotExistError, ProtocolEngineStoppedError
from ..types import EngineStatus
from .abstract_store import HasState, HandlesActions
from .actions import Action, UpdateCommandAction, PlayAction, PauseAction, StopAction


@dataclass(frozen=True)
class CommandState:
    """State of all protocol engine command resources."""

    is_running: bool
    stop_requested: bool
    # TODO(mc, 2021-06-16): OrderedDict is mutable. Switch to Sequence + Mapping
    commands_by_id: OrderedDict[str, Command]


class CommandStore(HasState[CommandState], HandlesActions):
    """Command state container."""

    _state: CommandState

    def __init__(self) -> None:
        """Initialize a CommandStore and its state."""
        self._state = CommandState(
            is_running=False,
            stop_requested=False,
            commands_by_id=OrderedDict(),
        )

    def handle_action(self, action: Action) -> None:
        """Modify state in reaction to an action."""
        if isinstance(action, UpdateCommandAction):
            command = action.command
            commands_by_id = self._state.commands_by_id.copy()
            commands_by_id.update({command.id: command})

            self._state = replace(self._state, commands_by_id=commands_by_id)

        elif isinstance(action, PlayAction):
            if not self._state.stop_requested:
                self._state = replace(self._state, is_running=True)

        elif isinstance(action, PauseAction):
            self._state = replace(self._state, is_running=False)

        elif isinstance(action, StopAction):
            self._state = replace(self._state, is_running=False, stop_requested=True)


class CommandView(HasState[CommandState]):
    """Read-only command state view."""

    _state: CommandState

    def __init__(self, state: CommandState) -> None:
        """Initialize the view of command state with its underlying data."""
        self._state = state

    def get(self, command_id: str) -> Command:
        """Get a command by its unique identifier."""
        try:
            return self._state.commands_by_id[command_id]
        except KeyError:
            raise CommandDoesNotExistError(f"Command {command_id} does not exist")

    def get_all(self) -> List[Command]:
        """Get a list of all commands in state.

        Entries are returned in the order of first-added command to last-added command.
        Replacing a command (to change its status, for example) keeps its place in the
        ordering.
        """
        return list(self._state.commands_by_id.values())

    def get_next_queued(self) -> Optional[str]:
        """Return the next request in line to be executed.

        Returns:
            The ID of the earliest queued command, if any.

        Raises:
            EngineStoppedError:
        """
        if self._state.stop_requested:
            raise ProtocolEngineStoppedError("Engine was stopped")

        if not self._state.is_running:
            return None

        for command_id, command in self._state.commands_by_id.items():
            if command.status == CommandStatus.FAILED:
                raise ProtocolEngineStoppedError("Previous command failed.")
            elif command.status == CommandStatus.QUEUED:
                return command_id

        return None

    def get_is_running(self) -> bool:
        """Get whether the engine is running and queued commands should be executed."""
        return self._state.is_running

    def get_is_complete(self, command_id: str) -> bool:
        """Get whether a given command is completed.

        A command is "completed" if one of the following is true:

        - Its status is CommandStatus.SUCCEEDED
        - Its status is CommandStatus.FAILED
        - A command earlier in the queue has a status of CommandStatus.FAILED
             - In this case, the command in question will never run

        Arguments:
            command_id: Command to check.
        """
        for search_id, search_command in self._state.commands_by_id.items():
            search_status = search_command.status
            is_failed = search_status == CommandStatus.FAILED

            if search_id == command_id or is_failed:
                return search_status == CommandStatus.SUCCEEDED or is_failed

        return False

    def get_all_complete(self) -> bool:
        """Get whether all commands have completed.

        All commands have "completed" if one of the following is true:

        - All commands have a status of CommandStatus.SUCCEEDED
        - Any command has a status of CommandStatus.FAILED
        """
        for command in self._state.commands_by_id.values():
            if command.status == CommandStatus.FAILED:
                return True
            elif command.status != CommandStatus.SUCCEEDED:
                return False
        return True

    def get_stop_requested(self) -> bool:
        """Get whether an engine stop has been requested.

        A command may still be executing while the engine is stopping.
        """
        return self._state.stop_requested

    def validate_action_allowed(self, action: Union[PlayAction, PauseAction]) -> None:
        """Validate if a PlayAction or PauseAction is allowed, raising if not.

        For safety / reliability reasons, a StopAction is always allowed.

        Raises:
            ProtocolEngineStoppedError: the engine has been stopped.
        """
        if self._state.stop_requested:
            action_desc = "play" if isinstance(action, PlayAction) else "pause"
            raise ProtocolEngineStoppedError(f"Cannot {action_desc} a stopped engine.")

    def get_status(self) -> EngineStatus:
        """Get the current execution status of the engine."""
        all_commands = self._state.commands_by_id.values()
        all_statuses = [c.status for c in all_commands]

        if self._state.stop_requested:
            if all(s == CommandStatus.SUCCEEDED for s in all_statuses):
                return EngineStatus.SUCCEEDED

            elif any(s == CommandStatus.RUNNING for s in all_statuses):
                return EngineStatus.STOP_REQUESTED

            else:
                return EngineStatus.STOPPED

        elif any(s == CommandStatus.FAILED for s in all_statuses):
            return EngineStatus.FAILED

        elif not self._state.is_running:
            if all(s == CommandStatus.QUEUED for s in all_statuses):
                return EngineStatus.READY_TO_RUN

            elif any(s == CommandStatus.RUNNING for s in all_statuses):
                return EngineStatus.PAUSE_REQUESTED

            else:
                return EngineStatus.PAUSED

        else:
            return EngineStatus.RUNNING



"""
- This driver is responsible for providing an interface for the mag deck
- The driver is the only system component that knows about the mag-deck's
  GCODES or how the mag-deck communicates

- The driver is NOT responsible for interpreting deck states in any way
  or knowing anything about what the device is being used for
"""

from __future__ import annotations

import asyncio
import logging
from typing import Dict, Optional
from enum import Enum

from opentrons.drivers import utils
from opentrons.drivers.command_builder import CommandBuilder
from opentrons.drivers.asyncio.communication import SerialConnection
from .abstract import AbstractMagDeckDriver

log = logging.getLogger(__name__)

DEFAULT_MAG_DECK_TIMEOUT = 10  # Quite large to account for probe time

DEFAULT_COMMAND_RETRIES = 3


class GCODE(str, Enum):
    HOME = "G28.2"
    PROBE_PLATE = "G38.2"
    GET_PLATE_HEIGHT = "M836"
    GET_CURRENT_POSITION = "M114.2"
    MOVE = "G0"
    DEVICE_INFO = "M115"
    PROGRAMMING_MODE = "dfu"


MAG_DECK_BAUDRATE = 115200

MAG_DECK_COMMAND_TERMINATOR = "\r\n\r\n"
MAG_DECK_ACK = "ok\r\nok\r\n"


# Number of digits after the decimal point for millimeter values
# being sent to/from magnetic module
GCODE_ROUNDING_PRECISION = 3


class MagDeckError(Exception):
    pass


class MagDeckDriver(AbstractMagDeckDriver):
    @classmethod
    async def create(
        cls, port: str, loop: Optional[asyncio.AbstractEventLoop] = None
    ) -> MagDeckDriver:
        """
        Create a mag deck driver.

        Args:
            port: port or url of magdeck
            loop: optional event loop

        Returns: driver
        """
        connection = await SerialConnection.create(
            port=port,
            baud_rate=MAG_DECK_BAUDRATE,
            timeout=DEFAULT_MAG_DECK_TIMEOUT,
            ack=MAG_DECK_ACK,
            loop=loop,
        )
        return cls(connection=connection)

    def __init__(self, connection: SerialConnection):
        """
        Constructor

        Args:
            connection: connection to magdeck.
        """
        self._connection = connection

    async def connect(self) -> None:
        """Connect to device"""
        await self._connection.open()

    async def disconnect(self) -> None:
        """Disconnect from device"""
        await self._connection.close()

    async def is_connected(self) -> bool:
        """Check if connected."""
        return await self._connection.is_open()

    async def home(self) -> None:
        """Homes the magnet"""
        c = CommandBuilder(terminator=MAG_DECK_COMMAND_TERMINATOR).add_gcode(
            gcode=GCODE.HOME
        )
        await self._send_command(c)

    async def probe_plate(self) -> None:
        """
        Probes for the deck plate and calculates the plate distance
        from home.
        To be used for calibrating MagDeck
        """
        c = CommandBuilder(terminator=MAG_DECK_COMMAND_TERMINATOR).add_gcode(
            gcode=GCODE.PROBE_PLATE
        )
        await self._send_command(c)

    async def get_plate_height(self) -> float:
        """
        Default plate_height for the device is 30;
        calculated as MAX_TRAVEL_DISTANCE(45mm) - 15mm
        """
        c = CommandBuilder(terminator=MAG_DECK_COMMAND_TERMINATOR).add_gcode(
            gcode=GCODE.GET_PLATE_HEIGHT
        )
        response = await self._send_command(c)
        data = utils.parse_key_values(response)
        return utils.parse_number(str(data.get("height")), GCODE_ROUNDING_PRECISION)

    async def get_mag_position(self) -> float:
        """
        Default mag_position for the device is 0.0
        i.e. it boots with the current position as 0.0
        """
        c = CommandBuilder(terminator=MAG_DECK_COMMAND_TERMINATOR).add_gcode(
            gcode=GCODE.GET_CURRENT_POSITION
        )

        response = await self._send_command(c)
        data = utils.parse_key_values(response)
        return utils.parse_number(str(data.get("Z")), GCODE_ROUNDING_PRECISION)

    async def move(self, position_mm: float) -> None:
        """
        Move the magnets along Z axis where the home position is 0.0;
        position_mm-> a point along Z. Does not self-check if the position
        is outside of the deck's linear range
        """
        c = (
            CommandBuilder(terminator=MAG_DECK_COMMAND_TERMINATOR)
            .add_gcode(gcode=GCODE.MOVE)
            .add_float(
                prefix="Z", value=position_mm, precision=GCODE_ROUNDING_PRECISION
            )
        )
        await self._send_command(c)

    async def get_device_info(self) -> Dict[str, str]:
        """
        Queries Mag-Deck for it's build version, model, and serial number

        Returns: dict
            Where keys are the strings 'version', 'model', and 'serial',
            and each value is a string identifier

            {
                'serial': '1aa11bb22',
                'model': '1aa11bb22',
                'version': '1aa11bb22'
            }

        Example input from Mag-Deck's serial response:
            "serial:aa11bb22 model:aa11bb22 version:aa11bb22"
        """
        c = CommandBuilder(terminator=MAG_DECK_COMMAND_TERMINATOR).add_gcode(
            gcode=GCODE.DEVICE_INFO
        )
        response = await self._send_command(c)
        return utils.parse_device_information(device_info_string=response)

    async def enter_programming_mode(self) -> None:
        """
        Enters and stays in DFU mode for 8 seconds.
        The module resets upon exiting the mode
        which causes the robot to lose serial connection to it.
        The connection can be restored by performing a .disconnect()
        followed by a .connect() to the same symlink node
        """
        c = CommandBuilder(terminator=MAG_DECK_COMMAND_TERMINATOR).add_gcode(
            gcode=GCODE.PROGRAMMING_MODE
        )
        await self._send_command(c)

    async def _send_command(self, command: CommandBuilder) -> str:
        """
        Send the command

        Args:
            command: command to send

        Returns:
            command response
        """
        response = await self._connection.send_command(
            command=command, retries=DEFAULT_COMMAND_RETRIES
        )
        return response
